from lona.html.abstract_node import AbstractNode
from lona.html.widget_data import WidgetData
from lona.html.node_list import NodeList
from lona.protocol import NODE_TYPE


class Widget(AbstractNode):
    FRONTEND_WIDGET_CLASS = ''

    @property
    def _id(self):
        if not hasattr(self, '_id_'):
            self._id_ = self.gen_id()

        return self._id_

    @property
    def nodes(self):
        if not hasattr(self, '_nodes'):
            self._nodes = NodeList(self)

        return self._nodes

    @nodes.setter
    def nodes(self, value):
        if not hasattr(self, '_nodes'):
            self._nodes = NodeList(self)

        self._nodes._reset(value)

    @property
    def data(self):
        if not hasattr(self, '_data'):
            self._data = WidgetData(self)

        return self._data

    @data.setter
    def data(self, value):
        if not hasattr(self, '_data'):
            self._data = WidgetData(self)

        self._data._reset(value)

    def __len__(self):
        return self._nodes.__len__()

    # node helper #############################################################
    def remove(self):
        if not self._parent:
            raise RuntimeError('node has no parent node')

        self._parent.remove(self)

    # serialisation ###########################################################
    def _has_changes(self):
        return self.nodes._has_changes() or self.data._has_changes()

    def _get_changes(self):
        return [
            *self.nodes._get_changes(),
            *self.data._get_changes(),
        ]

    def _clear_changes(self):
        self.nodes._clear_changes()
        self.data._clear_changes()

    def _serialize(self):
        return [
            NODE_TYPE.WIDGET,
            self._id,
            self.FRONTEND_WIDGET_CLASS,
            self.nodes._serialize(),
            self.data._serialize(),
        ]

    # event handling ##########################################################
    def on_click(self, input_event):
        return input_event

    def on_change(self, input_event):
        return input_event

    def on_submit(self, input_event):
        return input_event

    def handle_input_event(self, input_event):
        if input_event.name == 'click':
            return self.on_click(input_event)

        elif input_event.name == 'change':
            return self.on_change(input_event)

        elif input_event.name == 'submit':
            return self.on_submit(input_event)

        return input_event

    # string representation ###################################################
    def __str__(self):
        return '<!--lona-widget:{}-->\n{}\n<!--end-lona-widget:{}-->'.format(
            self._id,
            str(self.nodes),
            self._id,
        )

    def __repr__(self):
        return self.__str__()

    # node helper #############################################################
    def hide(self):
        with self.lock:
            for node in self.nodes:
                node.hide()

    def show(self):
        with self.lock:
            for node in self.nodes:
                node.show()

    def set_text(self, text):
        self.nodes = [str(text)]

    def get_text(self):
        with self.lock:
            text = []

            for node in self.nodes:
                text.append(node.get_text())

            return ' '.join(text).strip()



from concurrent.futures import ThreadPoolExecutor
import asyncio
import logging
import signal
import socket
import os

from aiohttp.web import Application, run_app

from lona.shell.shell import embed_shell, generate_shell_server
from lona.logging import LogFormatter, LogFilter
from lona.server import LonaServer

logger = logging.getLogger('lona')


def run_server(args):
    loop = asyncio.get_event_loop()

    # setup logging
    logging.basicConfig(level={
        'debug': logging.DEBUG,
        'info': logging.INFO,
        'warn': logging.WARN,
        'error': logging.ERROR,
        'critical': logging.CRITICAL,
    }[args.log_level.lower()])

    log_formatter = LogFormatter()
    log_filter = LogFilter()

    for handler in logging.getLogger().root.handlers:
        handler.setFormatter(log_formatter)
        handler.addFilter(log_filter)

    if args.loggers:
        for logger_name in args.loggers:
            if logger_name.startswith('_'):
                log_filter.exclude(logger_name[1:])

            else:
                if logger_name.startswith('+'):
                    logger_name = logger_name[1:]

                log_filter.include(logger_name)

    # setup lona server
    app = Application()

    server = LonaServer(
        app=app,
        project_root=args.project_root,
        settings_paths=args.settings,
        settings_pre_overrides=args.settings_pre_overrides,
        settings_post_overrides=args.settings_post_overrides,
    )

    executor = ThreadPoolExecutor(
        max_workers=server.settings.MAX_WORKERS,
        thread_name_prefix='LonaWorker',
    )

    async def shutdown(app):
        server = app['lona_server']

        await server.loop.run_in_executor(None, server.executor.shutdown)

    app.on_shutdown.append(shutdown)

    server.set_loop(loop)
    server.set_executor(executor)

    # run server
    if args.shell:
        async def start_shell(server):
            def _start_shell():
                embed_kwargs = {
                    'locals': {
                        'loop': loop,
                        'server': server,
                        'cli_args': vars(args),
                        'log_formatter': log_formatter,
                        'log_filter': log_filter,
                    },
                }

                embed_shell(server=server, **embed_kwargs)

                os.kill(os.getpid(), signal.SIGTERM)

            loop.run_in_executor(None, _start_shell)

        loop.create_task(start_shell(server))

    def _run_app():
        try:
            run_app(
                app=app,
                host=args.host,
                port=args.port,
                shutdown_timeout=args.shutdown_timeout,
            )

        except socket.gaierror as exception:
            if exception.errno not in (-2, ):
                raise

            logger.error('socket.gaierror: ' + exception.args[1])

        except OSError as exception:
            if exception.errno not in (13, 98):
                raise

            logger.error('OSError: ' + exception.args[1])

    if args.shell_server_url:
        embed_kwargs = {
            'bind': args.shell_server_url,
            'locals': {
                'loop': loop,
                'server': server,
                'cli_args': vars(args),
                'log_formatter': log_formatter,
                'log_filter': log_filter,
            },
        }

        repl_server = generate_shell_server(server=server, **embed_kwargs)

        with repl_server:
            _run_app()

    else:
        _run_app()



from lona.html import (
    TextInput as BaseTextInput,
    Button as BaseButton,
    Widget,
    Div,
    H5,
    P
)

from lona.static_files import StaticFile, StyleSheet, SORT_ORDER


STATIC_FILES = [
    # css files
    StyleSheet(
        name='bootstrap3.css',
        path='static/css/bootstrap.min.css',
        url='bootstrap3.min.css',
        sort_order=SORT_ORDER.FRAMEWORK,
    ),
    StyleSheet(
        name='bootstrap3-theme.css',
        path='static/css/bootstrap-theme.min.css',
        url='bootstrap3-theme.min.css',
        sort_order=SORT_ORDER.FRAMEWORK,
        enabled_by_default=False,
    ),

    # map files
    StaticFile(
        name='bootstrap3.css.map',
        path='static/css/bootstrap.min.css.map',
        url='bootstrap3.min.css.map',
        link=False,
    ),
    StaticFile(
        name='bootstrap3-theme.css.map',
        path='static/css/bootstrap-theme.min.css.map',
        url='bootstrap3-theme.min.css.map',
        link=False,
    ),

    # fonts
    StaticFile(
        name='bootstrap3-glyphicons-halflings-regular.eot',
        path='static/fonts/glyphicons-halflings-regular.eot',
        url='fonts/glyphicons-halflings-regular.eot',
        link=False,
    ),
    StaticFile(
        name='bootstrap3-glyphicons-halflings-regular.svg',
        path='static/fonts/glyphicons-halflings-regular.svg',
        url='fonts/glyphicons-halflings-regular.svg',
        link=False,
    ),
    StaticFile(
        name='bootstrap3-glyphicons-halflings-regular.ttf',
        path='static/fonts/glyphicons-halflings-regular.ttf',
        url='fonts/glyphicons-halflings-regular.ttf',
        link=False,
    ),
    StaticFile(
        name='bootstrap3-glyphicons-halflings-regular.woff',
        path='static/fonts/glyphicons-halflings-regular.woff',
        url='fonts/glyphicons-halflings-regular.woff',
        link=False,
    ),
    StaticFile(
        name='bootstrap3-glyphicons-halflings-regular.woff2',
        path='static/fonts/glyphicons-halflings-regular.woff2',
        url='fonts/glyphicons-halflings-regular.woff2',
        link=False,
    ),
]


# grid system #################################################################
class Row(Div):
    STATIC_FILES = STATIC_FILES
    CLASS_LIST = ['row']


class ColMd1(Div):
    STATIC_FILES = STATIC_FILES
    CLASS_LIST = ['col-md-1']


class ColMd2(Div):
    STATIC_FILES = STATIC_FILES
    CLASS_LIST = ['col-md-2']


class ColMd3(Div):
    STATIC_FILES = STATIC_FILES
    CLASS_LIST = ['col-md-3']


class ColMd4(Div):
    STATIC_FILES = STATIC_FILES
    CLASS_LIST = ['col-md-4']


class ColMd5(Div):
    STATIC_FILES = STATIC_FILES
    CLASS_LIST = ['col-md-5']


class ColMd6(Div):
    STATIC_FILES = STATIC_FILES
    CLASS_LIST = ['col-md-6']


class ColMd7(Div):
    STATIC_FILES = STATIC_FILES
    CLASS_LIST = ['col-md-7']


class ColMd8(Div):
    STATIC_FILES = STATIC_FILES
    CLASS_LIST = ['col-md-8']


class ColMd9(Div):
    STATIC_FILES = STATIC_FILES
    CLASS_LIST = ['col-md-9']


class ColMd10(Div):
    STATIC_FILES = STATIC_FILES
    CLASS_LIST = ['col-md-10']


class ColMd11(Div):
    STATIC_FILES = STATIC_FILES
    CLASS_LIST = ['col-md-11']


class ColMd12(Div):
    STATIC_FILES = STATIC_FILES
    CLASS_LIST = ['col-md-12']


# inputs ######################################################################
class TextInput(BaseTextInput):
    CLASS_LIST = ['form-control']


class InlineTextInput(TextInput):
    STYLE = {
        'width': 'auto',
        'display': 'inline',
    }


# buttons #####################################################################
class _Button(BaseButton):
    STATIC_FILES = STATIC_FILES

    ATTRIBUTES = {
        'type': 'button',
    }


class Button(_Button):
    STATIC_FILES = STATIC_FILES
    CLASS_LIST = ['btn', 'btn-primary']


class PrimaryButton(Button):
    STATIC_FILES = STATIC_FILES


class SecondaryButton(_Button):
    STATIC_FILES = STATIC_FILES
    CLASS_LIST = ['btn', 'btn-secondary']


class SuccessButton(_Button):
    STATIC_FILES = STATIC_FILES
    CLASS_LIST = ['btn', 'btn-success']


class DangerButton(_Button):
    STATIC_FILES = STATIC_FILES
    CLASS_LIST = ['btn', 'btn-danger']


class WarningButton(_Button):
    STATIC_FILES = STATIC_FILES
    CLASS_LIST = ['btn', 'btn-warning']


class InfoButton(_Button):
    STATIC_FILES = STATIC_FILES
    CLASS_LIST = ['btn', 'btn-info']


# popups ######################################################################
class ModalNode(Div):
    STATIC_FILES = STATIC_FILES
    CLASS_LIST = ['modal']

    ATTRIBUTES = {
        'tabindex': '-1',
        'role': 'dialog',
        'z-index': '2000',
    }


class ModalDialogNode(Div):
    STATIC_FILES = STATIC_FILES
    CLASS_LIST = ['modal-dialog']

    ATTRIBUTES = {
        'role': 'document',
    }


class ModalContentNode(Div):
    STATIC_FILES = STATIC_FILES
    CLASS_LIST = ['modal-content']


class ModalHeaderNode(Div):
    STATIC_FILES = STATIC_FILES
    CLASS_LIST = ['modal-header']


class ModalBodyNode(Div):
    STATIC_FILES = STATIC_FILES
    CLASS_LIST = ['modal-body']


class ModalFooterNode(Div):
    STATIC_FILES = STATIC_FILES
    CLASS_LIST = ['modal-footer']


class Modal(Widget):
    STATIC_FILES = STATIC_FILES

    class Mode:
        CONFIRM = 1
        CANCEL = 2

    def __init__(self, title='', body='', mode=Mode.CONFIRM):
        self.title = H5(title, _class='modal-title')
        self.body = P(body)

        self.buttons = []
        self.model_footer = ModalFooterNode()

        self.nodes = [ModalNode(
            ModalDialogNode(
                ModalContentNode(
                    ModalHeaderNode(
                        self.title,
                    ),
                    ModalBodyNode(
                        self.body,
                    ),
                    self.model_footer,
                ),
            ),
        )]

        self.set_mode(mode)

    def set_mode(self, mode):
        if mode == self.Mode.CONFIRM:
            self.buttons = [
                PrimaryButton('OK'),
                SecondaryButton('Cancel'),
            ]

            self.buttons[0].action = 'ok'
            self.buttons[1].action = 'cancel'

        elif mode == self.Mode.CANCEL:
            self.buttons = [
                DangerButton('Cancel'),
                SecondaryButton('Continue'),
            ]

            self.buttons[0].action = 'cancel'
            self.buttons[1].action = 'continue'

        else:
            raise ValueError('unknown mode')

        self.model_footer.nodes = self.buttons

    def set_title(self, title):
        self.title.set_text(title)

    def set_body(self, body):
        self.body.set_text(body)

    def show(self):
        self.nodes[0].style['display'] = 'block'

    def hide(self):
        self.nodes[0].style['display'] = 'none'


# progress bars ###############################################################
class ProgressBar(Widget):
    STATIC_FILES = STATIC_FILES

    def __init__(self, initial_value=None):
        self.progress_bar = Div(
            _class='progress-bar',
            _role='progress-bar',
            attributes={
                'aria-valuemin': '0',
                'aria-valuemax': '100',

            },
        )

        self.nodes = [Div(
            self.progress_bar,
            _class='progress',
        )]

        if initial_value:
            self.set_value(initial_value)

    def set_value(self, value, label=''):
        if isinstance(value, (tuple, list)):
            current, total = value

            if current == total:
                percentage = '100%'

            else:
                percentage = '{}%'.format(int((current / total) * 100))

            label = '{} ({}/{})'.format(percentage, current, total)

        else:
            percentage = '{}%'.format(value)

            if not label:
                label = percentage

        self.progress_bar.attributes['aria-valuenow'] = value
        self.progress_bar.style['width'] = percentage

        if label:
            self.progress_bar.set_text(label)



class LonaView:
    def handle_user_enter(self, request):
        pass

    def handle_request(self, *args, **kwargs):
        return ''

    def handle_input_event_root(self, input_event):
        return input_event

    def handle_input_event(self, input_event):
        return input_event



import threading
import asyncio

from lona.symbols import VIEW_RUNTIME_STATE
from lona.types import Symbol


class Client:
    def __init__(self, request):
        self.request = request

    def _assert_not_main_thread(self):
        if threading.current_thread() is threading.main_thread():
            raise RuntimeError(
                'this function is not supposed to run in the main thread')

    def _assert_view_is_interactive(self):
        if not self.request._view_runtime.route.interactive:
            raise RuntimeError(
                'operation is not supported in non-interactive requests')

    def _assert_view_is_running(self):
        if self.request._view_runtime.stop_reason:
            raise self.request._view_runtime.stop_reason

    def _await_specific_input_event(self, *nodes, event_type='', **kwargs):
        self.request._view_runtime.state = VIEW_RUNTIME_STATE.WAITING_FOR_INPUT

        try:
            self._assert_not_main_thread()
            self._assert_view_is_interactive()
            self._assert_view_is_running()

            if nodes:
                nodes = list(nodes)

            if len(nodes) == 1 and isinstance(nodes[0], list):
                nodes = nodes[0]

            if kwargs:
                self.show(**kwargs)

            return self.request._view_runtime.await_input_event(
                nodes=nodes,
                event_type=event_type,
            )

        finally:
            self.request._view_runtime.state = VIEW_RUNTIME_STATE.RUNNING

    def ping(self):
        self._assert_view_is_interactive()
        self._assert_view_is_running()

        return 'pong'

    def show(self, html=None, template=None, template_string=None, title=None,
             **kwargs):

        self._assert_not_main_thread()
        self._assert_view_is_interactive()
        self._assert_view_is_running()

        # templating
        if template or template_string:
            template_context = kwargs

            if 'template_context' in template_context:
                template_context = template_context['template_context']

            # string based templates
            if template_string:
                html = self.request.server.templating_engine.render_string(
                    template_string=template_string,
                    template_context=template_context,
                )

            # file based templates
            else:
                html = self.request.server.templating_engine.render_template(
                    template_name=template,
                    template_context=template_context,
                )

        with self.request._view_runtime.document.lock:
            html = html or self.request._view_runtime.document.html
            data = self.request._view_runtime.document.apply(html)

            if data:
                self.request._view_runtime.send_data(data=data, title=title)

    def set_title(self, title):
        self._assert_not_main_thread()
        self._assert_view_is_interactive()
        self._assert_view_is_running()

        with self.request._view_runtime.document.lock:
            self.request._view_runtime.send_data(title=title)

    def send_str(self, string, broadcast=False):
        self._assert_not_main_thread()
        self._assert_view_is_interactive()
        self._assert_view_is_running()

        if not broadcast:
            self.request.connection.send_str(string, sync=True)

            return

        for connection in self.request.server.websocket_connections:
            connection.send_str(string, sync=True)

    def await_input_event(self, **kwargs):
        return self. _await_specific_input_event(
            event_type='event',
            **kwargs,
        )

    def await_click(self, *clickable_nodes, **kwargs):
        return self. _await_specific_input_event(
            *clickable_nodes,
            event_type='click',
            **kwargs,
        )

    def await_change(self, *changeable_nodes, **kwargs):
        return self. _await_specific_input_event(
            *changeable_nodes,
            event_type='change',
            **kwargs,
        )

    def await_submit(self, *form_nodes, **kwargs):
        return self. _await_specific_input_event(
            *form_nodes,
            event_type='submit',
            **kwargs,
        )


class View:
    def __init__(self, request):
        self.request = request

    def daemonize(self):
        self.request.client._assert_view_is_interactive()

        self.request._view_runtime.is_daemon = True

    def _await_sync(self, awaitable):
        async def await_awaitable():
            finished, pending = await asyncio.wait(
                [
                    self.request._view_runtime.stopped,
                    awaitable,
                ],
                return_when=asyncio.FIRST_COMPLETED,
            )

            for finished_future in finished:
                if finished_future is not self.request._view_runtime.stopped:
                    return finished_future.result()

            if self.request._view_runtime.stopped in finished:
                for pending_future in pending:
                    if not pending_future.done():
                        pending_future.cancel()

                raise self.request._view_runtime.stop_reason

        return asyncio.run_coroutine_threadsafe(
            await_awaitable(),
            loop=self.request.server.loop,
        ).result()

    def await_sync(self, *args, **kwargs):
        self.request._view_runtime.state = \
            VIEW_RUNTIME_STATE.WAITING_FOR_IOLOOP

        try:
            return self._await_sync(*args, **kwargs)

        finally:
            self.request._view_runtime.state = VIEW_RUNTIME_STATE.RUNNING

    def sleep(self, *args, **kwargs):
        self.request._view_runtime.state = VIEW_RUNTIME_STATE.SLEEPING

        try:
            return self._await_sync(asyncio.sleep(*args, **kwargs))

        finally:
            self.request._view_runtime.state = VIEW_RUNTIME_STATE.RUNNING


class Request:
    def __init__(self, view_runtime, connection):
        self._view_runtime = view_runtime
        self.connection = connection

        self.url = self._view_runtime.url

        if self.url:
            self.GET = dict(self._view_runtime.url.query)
            self.POST = self._view_runtime.post_data or {}

        else:
            self.GET = {}
            self.POST = {}

        self.method = Symbol('POST' if self.POST else 'GET')

        self.server = self._view_runtime.server
        self.route = self._view_runtime.route
        self.match_info = self._view_runtime.match_info

        self.client = Client(self)
        self.view = View(self)

    @property
    def user(self):
        return getattr(self.connection, 'user', None)

    @property
    def frontend(self):
        return self._view_runtime.frontend



from html.parser import HTMLParser

from lona.html.text_node import TextNode
from lona.html.node import Node

SINGLE_TAGS = [
    'area',
    'base',
    'br',
    'col',
    'embed',
    'hr',
    'img',
    'input',
    'link',
    'meta',
    'param',
    'source',
    'track',
    'wbr',
]


class NodeHTMLParser(HTMLParser):
    def set_current_node(self, node):
        self._node = node

    def handle_starttag(self, tag, attrs):
        node_kwargs = {}

        # node attributes
        for name, value in attrs:
            if value is None:
                value = 'true'

            node_kwargs[name] = value

        # tag overrides
        node_kwargs['tag_name'] = tag
        node_kwargs['single_tag'] = tag in SINGLE_TAGS

        node = Node(**node_kwargs)
        self._node.append(node)
        self.set_current_node(node)

    def handle_endtag(self, tag):
        self.set_current_node(self._node.parent)

    def handle_data(self, data):
        self._node.append(TextNode(data))


def html_string_to_node_list(html_string):
    root_node = Node()
    html_parser = NodeHTMLParser()

    html_parser.set_current_node(root_node)
    html_parser.feed(html_string)

    return list(root_node.nodes)



import asyncio
import logging
import inspect

from lona.view import LonaView

logger = logging.getLogger('lona.view_loader')


class ViewLoader:
    def __init__(self, server):
        self.server = server

        self.setup()

    def _gen_cache_key(self, view):
        try:
            hash(view)

            return view

        except TypeError:
            return id(view)

    def _run_checks(self, route, view):
        # check if view is instance of lona.views.LonaView
        if(not route.http_pass_through and
           (not inspect.isclass(view) or not issubclass(view, LonaView))):

            logger.error('%s is no lona view', route.view)

            return

        # check if lona specific hooks are coroutine functions
        hook_names = [
            'handle_user_enter',
            'handle_request',
            'handle_input_event_root',
            'handle_input_event',
        ]

        for hook_name in hook_names:
            hook = getattr(view, hook_name, None)

            if(not route.http_pass_through and
               asyncio.iscoroutinefunction(hook)):

                logger.error(
                    '%s.%s is a coroutine function',
                    route.view,
                    hook_name,
                )

    def _generate_acquiring_error_view(self, exception):
        def view(request):
            raise exception

        return view

    def _acquire(self, view):
        logger.debug('loading %s', view)

        if isinstance(view, str):
            try:
                view = self.server.acquire(view)

            except Exception as exception:
                logger.error(
                    "exception raised while importing '%s'",
                    view,
                    exc_info=True,
                )

                view = self._generate_acquiring_error_view(exception)

        return view

    def setup(self):
        # views
        logger.debug('loading views from routes')

        self._cache = {}

        for route in self.server.router.routes:
            view = self._acquire(route.view)

            self._run_checks(route, view)

            cache_key = self._gen_cache_key(route.view)
            self._cache[cache_key] = view

            if route.frontend_view:
                view = self._acquire(route.frontend_view)
                cache_key = self._gen_cache_key(route.frontend_view)
                self._cache[cache_key] = view

        # special views
        import_strings = [
            # frontend
            self.server.settings.CORE_FRONTEND_VIEW,
            self.server.settings.FRONTEND_VIEW,

            # erorr 403
            self.server.settings.CORE_ERROR_403_VIEW,
            self.server.settings.ERROR_403_VIEW,

            # erorr 404
            self.server.settings.CORE_ERROR_404_VIEW,
            self.server.settings.ERROR_404_VIEW,

            # erorr 500
            self.server.settings.CORE_ERROR_500_VIEW,
            self.server.settings.ERROR_500_VIEW,
        ]

        for import_string in import_strings:
            if not import_string:
                continue

            # FIXME: run self._run_checks

            view = self._acquire(import_string)
            cache_key = self._gen_cache_key(import_string)
            self._cache[cache_key] = view

    def load(self, view):
        cache_key = self._gen_cache_key(view)

        return self._cache[cache_key]



from lona.static_files import StyleSheet, Script, SORT_ORDER
from lona.html import Widget, Canvas

STATIC_FILES = [
    # styesheets
    StyleSheet(
        name='chart_css_min',
        path='static/Chart.min.css',
        url='Chart.min.css',
        sort_order=SORT_ORDER.FRAMEWORK,
    ),
    StyleSheet(
        name='chart_css_min',
        path='static/Chart.css',
        url='Chart.min.css',
        sort_order=SORT_ORDER.FRAMEWORK,
        link=False,
    ),

    # scripts
    Script(
        name='chart_bundle_js_min',
        path='static/Chart.bundle.min.js',
        url='Chart.bundle.min.js',
        sort_order=SORT_ORDER.FRAMEWORK,
    ),
    Script(
        name='chart_bundle_js',
        path='static/Chart.bundle.js',
        url='Chart.bundle.js',
        sort_order=SORT_ORDER.FRAMEWORK,
        link=False,
    ),
    Script(
        name='chart_js_widget_js',
        path='static/chart-js-widget.js',
        url='chart-js-widget.js',
        sort_order=SORT_ORDER.LIBRARY,
    ),
]


class Chart(Widget):
    STATIC_FILES = STATIC_FILES
    FRONTEND_WIDGET_CLASS = 'chart_js'

    def __init__(self, **kwargs):
        self.nodes = [
            Canvas(),
        ]

        self.data = {
            **kwargs,
        }



class NotInteractiveError(Exception):
    pass


class Connection:
    def __init__(self, server, http_request, websocket=None):
        self.server = server
        self.http_request = http_request
        self.websocket = websocket

    @property
    def is_interactive(self):
        return self.websocket is not None

    @property
    def user(self):
        return getattr(self, '_user', None)

    @user.setter
    def user(self, value):
        self._user = value

    def send_str(self, string, sync=True):
        if not self.is_interactive:
            raise NotInteractiveError

        try:
            return self.server.run(
                self.websocket.send_str(string),
                sync=sync,
            )

        except ConnectionResetError:
            # this exception gets handled by aiohttp internally and
            # can be ignored

            pass



import importlib
import inspect
import runpy


def acquire(import_string, ignore_import_cache=False):
    attribute = None

    # scripts
    if '::' in import_string:
        script, attribute_name = import_string.split('::')
        attributes = runpy.run_path(script, run_name=script)

        if attribute_name not in attributes:
            raise ImportError("script '{}' has no attribute '{}'".format(
                script, attribute_name))

        attribute = attributes[attribute_name]

    # modules
    elif '.' in import_string:
        module_name, attribute_name = import_string.rsplit('.', 1)
        module = importlib.import_module(module_name)

        # ignore import cache
        if ignore_import_cache:
            path = inspect.getfile(module)

            return acquire('{}::{}'.format(path, attribute_name))

        if not hasattr(module, attribute_name):
            raise ImportError("module '{}' has no attribute '{}'".format(
                module_name, attribute_name))

        attribute = getattr(module, attribute_name)

    else:
        raise TypeError('invalid import string')

    return attribute



import asyncio
import logging
import inspect
import os

from aiohttp import WSMsgType

from aiohttp.web import (
    WebSocketResponse,
    FileResponse,
    HTTPFound,
    Response,
)

from lona.view_runtime_controller import ViewRuntimeController
from lona.middleware_controller import MiddlewareController
from lona.client_pre_compiler import ClientPreCompiler
from lona.response_parser import ResponseParser
from lona.static_files import StaticFileLoader
from lona.templating import TemplatingEngine
from lona.imports import acquire as _acquire
from lona.server_state import ServerState
from lona.view_runtime import ViewRuntime
from lona.view_loader import ViewLoader
from lona.connection import Connection
from lona.settings import Settings
from lona.routing import Router
from lona.types import Mapping

DEFAULT_SETTINGS = os.path.join(
    os.path.dirname(__file__),
    'default_settings.py',
)

server_logger = logging.getLogger('lona.server')
http_logger = logging.getLogger('lona.server.http')
websockets_logger = logging.getLogger('lona.server.websockets')


class LonaServer:
    def __init__(self, app, project_root, settings_paths=[],
                 settings_pre_overrides={}, settings_post_overrides={}):

        server_logger.debug("starting server in '%s'", project_root)

        self.project_root = os.path.abspath(project_root)
        self.websocket_connections = []
        self.user = Mapping()
        self._loop = None
        self._executor = None

        # setup aiohttp app
        self._app = app
        self._app['lona_server'] = self

        self._app.on_startup.append(self.start)
        self._app.on_shutdown.append(self.stop)

        # setup settings
        server_logger.debug('setup settings')

        self.settings_paths = [
            DEFAULT_SETTINGS,
        ]

        for path in settings_paths:
            self.settings_paths.append(
                os.path.normpath(
                    os.path.join(self.project_root, path),
                )
            )

        self.settings = Settings()

        if settings_pre_overrides:
            server_logger.debug('applying settings pre overrides')

            self.settings.update(settings_pre_overrides)

        for import_string in self.settings_paths:
            server_logger.debug("loading settings from '%s'", import_string)

            self.settings.add(import_string)

        if settings_post_overrides:
            server_logger.debug('applying settings post overrides')

            self.settings.update(settings_post_overrides)

        # setup server state
        server_logger.debug('setup server state')

        if self.settings.SERVER_STATE_ATOMIC:
            self._state = ServerState(initial_data={})

        else:
            self._state = {}

        # setup routing
        server_logger.debug('setup routing')
        self.router = Router()

        server_logger.debug("loading routing table from '%s'",
                            self.settings.ROUTING_TABLE)

        routes = self.acquire(self.settings.ROUTING_TABLE)

        if routes:
            self.router.add_routes(*routes)

        else:
            server_logger.warning('routing table is empty')

        # setup templating
        server_logger.debug('setup templating')

        self.templating_engine = TemplatingEngine(self)

        # setup websocket middleware controller
        server_logger.debug('setup middleware controller')

        self.middleware_controller = MiddlewareController(self)

        # setup aiohttp routes
        server_logger.debug('setup aiohttp routing')

        static_url = self.settings.STATIC_URL_PREFIX + '{path:.*}'

        server_logger.debug('static url set to %s', repr(static_url))

        self._app.router.add_route(
            '*', static_url, self._handle_static_file_request)

        self._app.router.add_route(
            '*', '/{path_info:.*}', self._handle_http_request)

        # setup view loader
        server_logger.debug('setup view loader')

        self.view_loader = ViewLoader(self)

        # setup response parser
        server_logger.debug('setup response parser')

        self.response_parser = ResponseParser(self)

        # setup views
        server_logger.debug('setup view runtime controller')

        self.view_runtime_controller = ViewRuntimeController(self)

        # setup static files
        self.client_pre_compiler = ClientPreCompiler(self)

        # the static file loader has to be started last because it does
        # node class discovery which has to happen after all views are imported
        server_logger.debug('setup static file')

        self.static_file_loader = StaticFileLoader(self)

        # setup startup hooks
        server_logger.debug('setup startup hooks')

        for hook in self.settings.STARTUP_HOOKS:
            if isinstance(hook, str):
                try:
                    hook = self.acquire(hook)

                except Exception:
                    server_logger.error(
                        "Exception occurred while importing startup hook '%s'",
                        hook,
                        exc_info=True,
                    )

                    continue

            self._app.on_startup.append(hook)

        # setup shutdown hooks
        server_logger.debug('setup shutdown hooks')

        for hook in self.settings.SHUTDOWN_HOOKS:
            if isinstance(hook, str):
                try:
                    hook = self.acquire(hook)

                except Exception:
                    server_logger.error(
                        "Exception occurred while importing shutdown hook '%s'",  # NOQA
                        hook,
                        exc_info=True,
                    )

                    continue

            self._app.on_shutdown.append(hook)

        # finish
        server_logger.debug('setup finish')

    def set_loop(self, loop):
        self._loop = loop

    def set_executor(self, executor):
        self._executor = executor

    @property
    def loop(self):
        return self._loop

    @property
    def executor(self):
        return self._executor

    async def start(self, *args, **kwargs):
        server_logger.debug('start')

        if not self._loop:
            raise RuntimeError('loop is not set')

        if not self._executor:
            raise RuntimeError('executor is not set')

        self.view_runtime_controller.start()

    async def stop(self, *args, **kwargs):
        server_logger.debug('stop')

        await self.run_function_async(self.view_runtime_controller.stop)

        for connection in self.websocket_connections.copy():
            try:
                await connection.websocket.close()

            except Exception:
                pass

    # asyncio helper ##########################################################
    def run_coroutine_sync(self, coroutine, wait=True):
        future = asyncio.run_coroutine_threadsafe(coroutine, loop=self._loop)

        if wait:
            return future.result()

        return future

    def run_function_async(self, function, *args, **kwargs):
        def _function():
            return function(*args, **kwargs)

        return self._loop.run_in_executor(self._executor, _function)

    def run(self, function_or_coroutine,
            *args, sync=False, wait=True, **kwargs):

        is_coroutine = asyncio.iscoroutine(function_or_coroutine)

        is_coroutine_function = asyncio.iscoroutinefunction(
            function_or_coroutine)

        # coroutine
        if is_coroutine or is_coroutine_function:
            coroutine = function_or_coroutine

            if is_coroutine_function:
                coroutine = function_or_coroutine(*args, **kwargs)

            # sync
            if sync:
                return self.run_coroutine_sync(coroutine, wait=wait)

            # async
            else:
                return coroutine

        # function
        else:
            function = function_or_coroutine

            def _function():
                return function(*args, **kwargs)

            # sync
            if sync:
                if wait:
                    return function(*args, **kwargs)

                return self._loop.run_in_executor(self._executor, _function)

            # async
            else:
                return self.run_function_async(function, *args, **kwargs)

    # path helper #############################################################
    def resolve_path(self, path):
        if path.startswith('/'):
            return path

        return os.path.normpath(os.path.join(self.project_root, path))

    def acquire(self, import_string, *args, **kwargs):
        if '::' in import_string:
            script, attribute_name = import_string.split('::')
            script = self.resolve_path(script)

            import_string = '{}::{}'.format(script, attribute_name)

        return _acquire(import_string, *args, **kwargs)

    # connection management ###################################################
    async def _setup_connection(self, http_request, websocket=None):
        connection = Connection(self, http_request, websocket)

        handled, data, middleware = \
            await self.middleware_controller.process_connection(
                connection,
            )

        if websocket is not None:
            self.websocket_connections.append(connection)

            if connection.user not in self.user:
                self.user[connection.user] = []

            self.user[connection.user].append(connection)

        return connection, (handled, data, middleware)

    def _remove_connection(self, connection):
        if connection in self.websocket_connections:
            self.websocket_connections.remove(connection)

        if connection.user in self.user:
            self.user[connection.user].remove(connection)

            if len(self.user[connection.user]) == 0:
                self.user.pop(connection.user)

    # view helper #############################################################
    def _render_response(self, response_dict):
        if response_dict['file']:
            return FileResponse(response_dict['file'])

        if response_dict['redirect']:
            return HTTPFound(response_dict['redirect'])

        if response_dict['http_redirect']:
            return HTTPFound(response_dict['http_redirect'])

        response = Response(
            status=response_dict['status'],
            content_type=response_dict['content_type'],
            text=response_dict['text'],
        )

        response.headers['Cache-Control'] = 'no-cache, no-store, must-revalidate'  # NOQA

        return response

    # handle http requests ####################################################
    async def _handle_static_file_request(self, request):
        def _404():
            return Response(
                status=404,
                text='404: Not found',
            )

        if not self.settings.STATIC_FILES_SERVE:
            server_logger.warning(
                'Reverse proxy seems to be misconfigured: a static file request was received but STATIC_FILES_SERVE is disabled',  # NOQA
            )

            return _404()

        rel_path = request.match_info['path']

        abs_path = await self.run_function_async(
            self.static_file_loader.resolve_path,
            rel_path,
        )

        if not abs_path:
            return _404()

        return FileResponse(abs_path)

    async def _handle_websocket_message(self, connection, message):
        websockets_logger.debug(
            '%s message received %s', connection, message.data)

        handled, data, middleware = \
            await self.middleware_controller.process_websocket_message(
                connection,
                message.data,
            )

        if not handled:
            websockets_logger.debug(
                '%s message got not handled', connection)

    async def _handle_websocket_request(self, http_request):
        websocket = None
        connection = None

        async def close_websocket():
            self.view_runtime_controller.remove_connection(connection)
            self._remove_connection(connection)

            await websocket.close()

            websockets_logger.debug('%s closed', connection)

        # setup websocket
        websocket = WebSocketResponse()
        await websocket.prepare(http_request)

        # setup connection
        connection, middleware_data = await self._setup_connection(
            http_request,
            websocket,
        )

        handled, data, middleware = middleware_data

        # connection got closed by middleware
        if handled:
            if data:
                if not isinstance(data, str):
                    raise RuntimeError('%s.process_connection returned non string data'.format(middleware))  # NOQA

                await connection.send_str(data, sync=False)

            await close_websocket()

            return websocket

        websockets_logger.debug('%s opened', connection)

        # main loop
        try:
            async for message in websocket:
                if message.type == WSMsgType.TEXT:
                    self._loop.create_task(
                        self._handle_websocket_message(connection, message),
                    )

                elif message.type == WSMsgType.PING:
                    await websocket.pong()

                elif message.type in (WSMsgType.CLOSED, WSMsgType.ERROR):
                    break

        except asyncio.CancelledError:
            websockets_logger.debug('CancelledError')

        finally:
            await close_websocket()

        return websocket

    async def _handle_http_request(self, http_request):
        http_logger.debug('http request incoming')

        # resolve path
        http_logger.debug("resolving path %s", repr(http_request.path))

        match, route, match_info = await self.run_function_async(
            self.router.resolve,
            http_request.path,
        )

        if match:
            http_logger.debug('route %s matched', route)

        else:
            http_logger.debug('no route matched')

        # http pass through
        # FIXME: add support for handle_user_enter
        if match and route.http_pass_through:
            http_logger.debug('http_pass_through mode')

            # load view
            view = self.view_loader.load(route.view)

            if inspect.isclass(view):
                view = view().handle_request

            # run view
            if asyncio.iscoroutinefunction(view):
                response = await view(http_request)

            else:
                response = await self.run_function_async(
                    view,
                    http_request,
                )

            if asyncio.iscoroutine(response):
                response = await response

            # render and return response
            if isinstance(response, dict):
                return await self.run_function_async(
                    self._render_response,
                    response,
                )

            return response

        # websocket requests
        if(http_request.method == 'GET' and
           http_request.headers.get('upgrade', '').lower() == 'websocket'):

            return await self._handle_websocket_request(http_request)

        # setup connection
        try:
            connection, middleware_data = await self._setup_connection(
                http_request,
            )

        except Exception:
            http_logger.error(
                'Exception occurred while setting connection up',
                exc_info=True,
            )

            return Response(status=500, body='500: Internal Error')

        handled, data, middleware = middleware_data

        # connection got closed by middleware
        if handled:
            if data:
                return self._render_response(data)

            return Response(status=503, body='503: Service Unavailable')

        # run non interactive view or frontend
        if match and not route.interactive:
            frontend = False

            http_logger.debug('non-interactive mode')

        else:
            http_logger.debug('frontend mode')

            frontend = True

        view_runtime = ViewRuntime(
            server=self,
            url=http_request.path,
            route=route,
            match_info=match_info,
            post_data=await http_request.post(),
            frontend=frontend,
            start_connection=connection,
        )

        response_dict = await self.run_function_async(
            view_runtime.run_middlewares,
        )

        if not response_dict:
            response_dict = await self.run_function_async(
                view_runtime.start,
            )

        return self._render_response(response_dict)

    # state ###################################################################
    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, data):
        if self.settings.SERVER_STATE_ATOMIC:
            self._state._reset(data)

        else:
            self._state = data

    # helper ##################################################################
    def get_running_views_count(self, *args, **kwargs):
        return self.view_runtime_controller.get_running_views_count(
            *args,
            **kwargs,
        )

    def view_is_already_running(self, *args, **kwargs):
        return self.view_runtime_controller.view_is_already_running(
            *args,
            **kwargs,
        )

    def get_connection_count(self, user):
        if user not in self.user:
            return 0

        return len(self.user[user])

    def get_template(self, *args, **kwargs):
        return self.templating_engine.get_template(*args, **kwargs)

    def render_string(self, *args, **kwargs):
        return self.templating_engine.render_string(*args, **kwargs)

    def render_template(self, *args, **kwargs):
        return self.templating_engine.render_template(*args, **kwargs)

    def reverse(self, *args, **kwargs):
        return self.router.reverse(*args, **kwargs)



from lona.events.event_types import CLICK
from lona.html.widget import Widget  # NOQA
from lona.html.node import Node


# simple html nodes ###########################################################
class Html(Node):
    TAG_NAME = 'html'


class Head(Node):
    TAG_NAME = 'head'


class Body(Node):
    TAG_NAME = 'body'


class Nav(Node):
    TAG_NAME = 'nav'


class Main(Node):
    TAG_NAME = 'main'


class Section(Node):
    TAG_NAME = 'section'


class Script(Node):
    TAG_NAME = 'script'


class NoScript(Node):
    TAG_NAME = 'noscript'


class H1(Node):
    TAG_NAME = 'h1'


class H2(Node):
    TAG_NAME = 'h2'


class H3(Node):
    TAG_NAME = 'h3'


class H4(Node):
    TAG_NAME = 'h4'


class H5(Node):
    TAG_NAME = 'h5'


class H6(Node):
    TAG_NAME = 'h6'


class Div(Node):
    TAG_NAME = 'div'


class Span(Node):
    TAG_NAME = 'span'


class P(Node):
    TAG_NAME = 'p'


class I(Node):  # NOQA
    TAG_NAME = 'i'


class Strong(Node):
    TAG_NAME = 'strong'


class Small(Node):
    TAG_NAME = 'small'


class Ul(Node):
    TAG_NAME = 'ul'


class Ol(Node):
    TAG_NAME = 'ol'


class Li(Node):
    TAG_NAME = 'li'


class Table(Node):
    TAG_NAME = 'table'


class THead(Node):
    TAG_NAME = 'thead'


class TBody(Node):
    TAG_NAME = 'tbody'


class TFood(Node):
    TAG_NAME = 'tfood'


class ColGroup(Node):
    TAG_NAME = 'colgroup'


class Tr(Node):
    TAG_NAME = 'tr'


class Th(Node):
    TAG_NAME = 'th'


class Td(Node):
    TAG_NAME = 'td'


class Canvas(Node):
    TAG_NAME = 'canvas'


class Hr(Node):
    TAG_NAME = 'hr'
    SINGLE_TAG = True


class Br(Node):
    TAG_NAME = 'br'
    SINGLE_TAG = True


# inputs ######################################################################
class Form(Node):
    TAG_NAME = 'form'


class Fieldset(Node):
    TAG_NAME = 'form'


class Label(Node):
    TAG_NAME = 'label'


class Button(Node):
    TAG_NAME = 'button'
    EVENTS = [CLICK]


class InputNode(Node):
    TAG_NAME = 'input'
    SINGLE_TAG = True


class Submit(InputNode):
    ATTRIBUTES = {
        'type': 'submit',
        'value': 'Submit',
    }


class Reset(InputNode):
    ATTRIBUTES = {
        'type': 'reset',
        'value': 'Reset',
    }


class TextInputNode(InputNode):
    ATTRIBUTES = {
        'type': 'text',
    }


class CheckboxNode(InputNode):
    ATTRIBUTES = {
        'type': 'checkbox',
    }


class TextAreaNode(Node):
    TAG_NAME = 'textarea'

    def __repr__(self):
        return self.__str__(
            skip_value=True,
            node_string=self.attributes.get('value', ''),
        )


class SelectNode(Node):
    TAG_NAME = 'select'


class OptionNode(Node):
    TAG_NAME = 'option'


# complex html nodes ##########################################################
class Img(Node):
    TAG_NAME = 'img'
    SINGLE_TAG = True


class A(Node):
    TAG_NAME = 'a'

    def __init__(self, *args, interactive=None, **kwargs):
        super().__init__(*args, **kwargs)

        if interactive is not None:
            self.interactive = interactive

    @property
    def interactive(self):
        return self.ignore

    @interactive.setter
    def interactive(self, value):
        self.ignore = value

    def set_href(self, href):
        self.attributes['href'] = str(href)

    def get_href(self, href):
        return self.attributes['href']


class Pre(Node):
    TAG_NAME = 'pre'

    def write(self, string):
        string = str(string)

        if len(self.nodes) > 0 and isinstance(self.nodes[-1], str):
            self.nodes[-1] = self.nodes[-1] + string

        else:
            self.nodes.append(string)

    def write_line(self, string):
        string = str(string) + '\n'

        self.nodes.append(string)



from time import monotonic_ns

from lona.protocol import OPERATION, PATCH_TYPE


class AttributeList:
    PATCH_TYPE = None

    def __init__(self, node, *args, **kwargs):
        self._node = node
        self._attributes = list(*args, **kwargs)
        self._changes = []

    # list helper #############################################################
    def add(self, attribute):
        if not isinstance(attribute, (int, bool, float, str)):
            raise ValueError('unsupported type')

        with self._node.document.lock:
            if attribute in self._attributes:
                return

            self._attributes.append(attribute)

            self._changes.append([
                monotonic_ns(),
                self._node._id,
                self.PATCH_TYPE,
                OPERATION.ADD,
                attribute,
            ])

    def remove(self, attribute):
        with self._node.document.lock:
            if attribute not in self._attributes:
                return

            self._attributes.remove(attribute)

            self._changes.append([
                monotonic_ns(),
                self._node._id,
                self.PATCH_TYPE,
                OPERATION.REMOVE,
                attribute,
            ])

    def clear(self):
        with self._node.document.lock:
            if not self._attributes:
                return

            self._attributes.clear()

            self._changes.append([
                monotonic_ns(),
                self._node._id,
                self.PATCH_TYPE,
                OPERATION.CLEAR,
            ])

    def toggle(self, attribute):
        with self._node.document.lock:
            if attribute in self._attributes:
                self.remove(attribute)

            else:
                self.add(attribute)

    def append(self, attribute):
        return self.add(attribute)

    def extend(self, attributes):
        with self._node.document.lock:
            for attribute in iter(attributes):
                if attribute in self._attributes:
                    continue

                self.append(attribute)

    def __len__(self):
        with self._node.document.lock:
            return len(self._attributes)

    def __bool__(self):
        with self._node.document.lock:
            return bool(self._attributes)

    def __iter__(self):
        with self._node.document.lock:
            return self._attributes.__iter__()

    # serialisation ###########################################################
    def _reset(self, value):
        if not isinstance(value, list):
            raise ValueError('unsupported type')

        for i in value:
            if not isinstance(i, (int, bool, float, str)):
                raise ValueError('unsupported type')

        with self._node.document.lock:
            self._attributes = value

            self._changes.append([
                monotonic_ns(),
                self._node._id,
                self.PATCH_TYPE,
                OPERATION.RESET,
                list(value),
            ])

    def _has_changes(self):
        return bool(self._changes)

    def _get_changes(self):
        return list(self._changes)

    def _clear_changes(self):
        return self._changes.clear()

    def _serialize(self):
        return list(self._attributes)

    # string representation ###################################################
    def __str__(self):
        with self._node.document.lock:
            return ' '.join([i for i in self._attributes])

    def __repr__(self):
        return '<AttributeList({})>'.format(repr(self._attributes))


class IDList(AttributeList):
    PATCH_TYPE = PATCH_TYPE.ID_LIST

    def __repr__(self):
        return '<IDList({})>'.format(repr(self._attributes))


class ClassList(AttributeList):
    PATCH_TYPE = PATCH_TYPE.CLASS_LIST

    def __repr__(self):
        return '<ClassList({})>'.format(repr(self._attributes))



from lona.protocol import INPUT_EVENT_TYPE


class InputEvent:
    def __init__(self, request, payload, document):
        self.request = request
        self.payload = payload
        self.document = document

        self.data = {}
        self.node = None
        self.widgets = []
        self.tag_name = ''
        self.id_list = []
        self.class_list = []

        # parse input event type
        if isinstance(payload[0], str):
            self.type = INPUT_EVENT_TYPE.CUSTOM
            self.name = payload[0]
            self.data = payload[1]
            self.node_info = payload[2:]

        elif payload[0] == INPUT_EVENT_TYPE.CLICK:
            self.type = INPUT_EVENT_TYPE.CLICK
            self.name = 'click'
            self.data = payload[1]
            self.node_info = payload[2:]

        elif payload[0] == INPUT_EVENT_TYPE.CHANGE:
            self.type = INPUT_EVENT_TYPE.CHANGE
            self.name = 'change'
            self.data = payload[1]
            self.node_info = payload[2:]

        elif payload[0] == INPUT_EVENT_TYPE.SUBMIT:
            self.type = INPUT_EVENT_TYPE.SUBMIT
            self.name = 'submit'
            self.data = payload[1]
            self.node_info = payload[2:]

        # find node
        # node info contains a lona node id
        if self.node_info[0] or self.node_info[1]:
            self.node, self.widgets = document.get_node(
                widget_id=self.node_info[0],
                node_id=self.node_info[1],
            )

        self.tag_name = self.node_info[2]
        self.id_list = (self.node_info[3] or '').split(' ')
        self.class_list = (self.node_info[4] or '').split(' ')

    def node_has_id(self, name):
        if self.node is None:
            return name in self.id_list

        return self.node.has_id(name)

    def node_has_class(self, name):
        if self.node is None:
            return name in self.class_list

        return self.node.has_class(name)



from textwrap import indent

from lona.html.attribute_dict import AttributeDict, StyleDict
from lona.html.attribute_list import IDList, ClassList
from lona.html.node_event_list import NodeEventList
from lona.html.abstract_node import AbstractNode
from lona.html.node_list import NodeList
from lona.protocol import NODE_TYPE


class Node(AbstractNode):
    TAG_NAME = 'html'
    SINGLE_TAG = False
    ID_LIST = []
    CLASS_LIST = []
    STYLE = {}
    ATTRIBUTES = {}
    EVENTS = []

    def __init__(self, *args, tag_name=None, single_tag=None, **kwargs):
        self._id = self.gen_id()

        self._id_list = IDList(self, self.ID_LIST)
        self._class_list = ClassList(self, self.CLASS_LIST)
        self._style = StyleDict(self, self.STYLE)
        self._attributes = AttributeDict(self, self.ATTRIBUTES)
        self._nodes = NodeList(self)
        self._events = NodeEventList(self, self.EVENTS)

        # tag overrides
        self.tag_name = tag_name or self.TAG_NAME

        if single_tag is None:
            self.single_tag = self.SINGLE_TAG

        else:
            self.single_tag = single_tag

        # args (nodes)
        for arg in args:
            self.append(arg)

        # kwargs (attributes)
        for name, value in kwargs.items():

            # remove underscores from attributes
            # this makes kwargs like '_class' possible to prevent clashes
            # with python grammar
            if '_' in name:
                name = name.replace('_', '-')

                if name.startswith('-'):
                    name = name[1:]

            # lona classes
            if name in ('ignore'):
                setattr(self, name, value)

            # change aware attributes
            elif name == 'id':
                if not isinstance(value, (str, list)):
                    raise ValueError('id has to be string or list of srings')  # NOQA

                if isinstance(value, str):
                    value = value.split(' ')

                self._id_list.extend(value)

            elif name == 'class':
                if not isinstance(value, (str, list)):
                    raise ValueError('class has to be string or list of srings')  # NOQA

                if isinstance(value, str):
                    value = value.split(' ')

                self._class_list.extend(value)

            elif name == 'style':
                if not isinstance(value, dict):
                    raise ValueError('style has to be dict')

                self._style.update(value)

            elif name == 'attributes':
                if not isinstance(value, dict):
                    raise ValueError('attributes has to be dict')

                self._attributes.update(value)

            elif name == 'events':
                if not isinstance(value, list):
                    raise ValueError('events have to be list')

                self._events.extend(value)

            elif name == 'nodes':
                if not isinstance(value, list):
                    value = [value]

                self._nodes._reset(value)

            # misc attributes
            else:
                self._attributes[name] = value

    # node attributes  ########################################################
    # id_list
    @property
    def id_list(self):
        return self._id_list

    @id_list.setter
    def id_list(self, value):
        self._id_list._reset(value)

    # class_list
    @property
    def class_list(self):
        return self._class_list

    @class_list.setter
    def class_list(self, value):
        self._class_list._reset(value)

    # style
    @property
    def style(self):
        return self._style

    @style.setter
    def style(self, value):
        self._style._reset(value)

    # attributes
    @property
    def attributes(self):
        return self._attributes

    @attributes.setter
    def attributes(self, value):
        self._attributes._reset(value)

    # events
    @property
    def events(self):
        return self._events

    @events.setter
    def events(self, value):
        self._events._reset(value)

    # nodes
    @property
    def nodes(self):
        return self._nodes

    @nodes.setter
    def nodes(self, value):
        self._nodes._reset(value)

    # lona attribute helper ###################################################
    def has_class(self, class_name):
        return class_name in self._class_list

    def has_id(self, id_name):
        return id_name in self._id_list

    # lona attributes #########################################################
    @property
    def ignore(self):
        return 'data-lona-ignore' in self._attributes

    @ignore.setter
    def ignore(self, value):
        if value:
            self._attributes['data-lona-ignore'] = 'True'

        else:
            with self.lock:
                if 'data-lona-ignore' in self._attributes:
                    del self._attributes['data-lona-ignore']

    # serialisation ###########################################################
    def _has_changes(self):
        return any([
            self._id_list._has_changes(),
            self._class_list._has_changes(),
            self._style._has_changes(),
            self._attributes._has_changes(),
            self._nodes._has_changes(),
        ])

    def _get_changes(self):
        return [
            *self._id_list._get_changes(),
            *self._class_list._get_changes(),
            *self._style._get_changes(),
            *self._attributes._get_changes(),
            *self._nodes._get_changes(),
        ]

    def _clear_changes(self):
        self._id_list._clear_changes()
        self._class_list._clear_changes()
        self._style._clear_changes()
        self._attributes._clear_changes()
        self._nodes._clear_changes()

    def _serialize(self):
        return [
            NODE_TYPE.NODE,
            self._id,
            self.tag_name,
            self._id_list._serialize(),
            self._class_list._serialize(),
            self._style._serialize(),
            self._attributes._serialize(),
            self._nodes._serialize(),
        ]

    # node list helper ########################################################
    def insert(self, index, node):
        self._nodes.insert(index, node)

    def append(self, node):
        self._nodes.append(node)

    def remove(self, node=None):
        if not node:
            if not self._parent:
                raise RuntimeError('node has no parent node')

            self._parent.remove(self)

            return

        self._nodes.remove(node)

    def pop(self, index):
        return self._nodes.pop(index)

    def clear(self):
        self._nodes.clear()

    def __getitem__(self, index):
        return self._nodes[index]

    def __setitem__(self, index, value):
        self._nodes[index] = value

    def __iter__(self):
        return self._nodes.__iter__()

    def __len__(self):
        return self._nodes.__len__()

    def __bool__(self):
        return True

    # string representation ###################################################
    def __str__(self, node_string=None, skip_value=False):
        with self.document.lock:
            # opening tag
            string = '<{} data-lona-node-id="{}"'.format(
                self.tag_name,
                self._id,
            )

            if self.id_list:
                string += ' id="{}"'.format(str(self.id_list))

            if self.class_list:
                string += ' class="{}"'.format(str(self.class_list))

            if self.style:
                string += ' style="{}"'.format(
                    self.style.to_sub_attribute_string())

            if self.attributes:
                string += ' '

                string += self.attributes.to_attribute_string(
                    skip_value=skip_value,
                )

            if self.single_tag:
                string += ' />'

            else:
                string += '>'

            # nodes
            if node_string:
                string += '\n{}\n'.format(
                    indent(node_string, '  '),
                )

            elif self.nodes:
                string += '\n'
                string += indent(str(self.nodes), '  ')
                string += '\n'

            # closing tag
            if not self.single_tag:
                string += '</{}>'.format(self.tag_name)

            return string

    def __repr__(self):
        return self.__str__()

    # HTML helper #############################################################
    def set_text(self, text):
        self.nodes = [str(text)]

    def get_text(self):
        with self.lock:
            text = []

            for node in self.nodes:
                text.append(node.get_text())

            text = ' '.join(text)

            # strip starting and ending whitespaces but preserve \n
            # this is important for nodes like <pre> which are newline aware
            while text.startswith('\t') or text.startswith(' '):
                text = text[1:]

            while text.endswith('\t') or text.endswith(' '):
                text = text[:1]

            return text

    def hide(self):
        self.style['display'] = 'none'

    def show(self):
        with self.lock:
            if 'display' in self.style and self.style['display'] == 'none':
                del self.style['display']



import logging

from yarl import URL

from lona.protocol import encode_http_redirect, METHOD
from lona.view_runtime import ViewRuntime
from lona.exceptions import ServerStop
from lona.types import Mapping

logger = logging.getLogger('lona.view_runtime_controller')


class ViewRuntimeController:
    def __init__(self, server):
        self.server = server

        self.running_single_user_views = Mapping()
        # contains: {
        #     connection.user: [
        #         view_runtime,
        #     ]
        # }

        self.running_multi_user_views = Mapping()
        # contains: {
        #    route: view_runtime,
        # }

    def start(self):
        return

    def stop(self):
        # running views per user
        for user, view_runtimes in self.running_single_user_views.items():
            for view_runtime in view_runtimes.copy():
                view_runtime.stop(reason=ServerStop)

        # multi user views
        for route, view in self.running_multi_user_views.items():
            view.stop(reason=ServerStop)

    # helper ##################################################################
    def iter_single_user_view_runtimes(self):
        # this method is not thread safe
        # only for debug purposes
        # yields: view_runtime

        running_single_user_views = self.running_single_user_views.copy()

        for user, view_runtimes in running_single_user_views.items():
            for view_runtime in view_runtimes.copy():
                yield view_runtime

    def iter_multi_user_view_runtimes(self):
        # this method is not thread safe
        # only for debug purposes
        # yields: view_runtime

        running_multi_user_views = self.running_multi_user_views.copy()

        for route, view_runtime in running_multi_user_views.items():
            yield view_runtime

    def get_view_runtime(self, view_runtime_id):
        # this method is not thread safe
        # only for debug purposes
        # yields: view_runtime

        # multi user view runtimes
        for view_runtime in self.iter_multi_user_view_runtimes():
            if view_runtime.view_runtime_id == view_runtime_id:
                return view_runtime

        # single user view runtimes
        for view_runtime in self.iter_single_user_view_runtimes():
            if view_runtime.view_runtime_id == view_runtime_id:
                return view_runtime

    def get_running_views_count(self, user):
        if user not in self.running_single_user_views:
            return 0

        count = 0

        for view_runtime in self.running_single_user_views[user]:
            if not view_runtime.frontend and not view_runtime.is_stopped:
                count += 1

        return count

    def view_is_already_running(self, request):
        if request.user in self.running_single_user_views:
            for view_runtime in self.running_single_user_views[request.user]:
                if(view_runtime.route == request.route and
                   view_runtime.match_info == request.match_info and
                   view_runtime.is_daemon):

                    return True

        return False

    # view management #########################################################
    def remove_connection(self, connection, window_id=None):
        for user, view_runtimes in self.running_single_user_views.items():
            for view_runtime in view_runtimes:
                view_runtime.remove_connection(connection, window_id=None)

        for route, view in self.running_multi_user_views.items():
            view.remove_connection(connection, window_id=None)

    def remove_view_runtime(self, view_runtime):
        user = view_runtime.start_connection.user

        self.running_single_user_views[user].remove(view_runtime)

    # lona messages ###########################################################
    def handle_lona_message(self, connection, window_id, view_runtime_id,
                            method, payload):

        """
        this method gets called by the
        lona.middlewares.LonaMessageMiddleware.process_websocket_message

        """

        # views
        if method == METHOD.VIEW:
            url, post_data = payload

            url_object = URL(url)

            # disconnect client window from previous view
            self.remove_connection(connection, window_id)

            # resolve url
            match, route, match_info = self.server.router.resolve(
                url_object.path)

            # route is not interactive; issue a http redirect
            if match and (route.http_pass_through or not route.interactive):
                message = encode_http_redirect(
                    window_id=window_id,
                    view_runtime_id=None,
                    target_url=url,
                )

                connection.send_str(message)

                return

            view_runtime = ViewRuntime(
                server=self.server,
                url=url,
                route=route,
                match_info=match_info,
                post_data=post_data or {},
                start_connection=connection,
            )

            view_runtime.add_connection(
                connection=connection,
                window_id=window_id,
                url=url,
            )

            response_dict = view_runtime.run_middlewares()

            # request got handled by a middleware
            if response_dict:
                return

            # check for 403 forbidden error
            response_dict = view_runtime.run_user_enter()

            if response_dict:
                return

            # reconnect or close previous started single user views
            # for this route
            user = connection.user
            running_view_runtime = None

            if user in self.running_single_user_views:
                for _view_runtime in self.running_single_user_views[user]:
                    if(_view_runtime.route == route and
                       _view_runtime.match_info == match_info and
                       _view_runtime.is_daemon):

                        running_view_runtime = _view_runtime

                        break

            if running_view_runtime:
                if not running_view_runtime.is_stopped:
                    running_view_runtime.reconnect_connection(
                        connection=connection,
                        window_id=window_id,
                        url=url,
                    )

                    return

                else:
                    running_view_runtime.stop()

            # connect to a multi user view
            elif(route in self.running_multi_user_views):
                self.running_multi_user_views[route].reconnect_connection(
                    connection=connection,
                    window_id=window_id,
                    url=url,
                )

                return

            # start view
            if user not in self.running_single_user_views:
                self.running_single_user_views[user] = []

            self.running_single_user_views[user].append(view_runtime)

            view_runtime.start()

        # input events
        elif method == METHOD.INPUT_EVENT:
            user = connection.user

            if user not in self.running_single_user_views:
                return

            for view_runtime in self.running_single_user_views[user]:
                if view_runtime.view_runtime_id == view_runtime_id:
                    view_runtime.handle_input_event(connection, payload)

                    break



from copy import deepcopy, copy

from lona.html.parsing import html_string_to_node_list
from lona.html import Widget, CHANGE


class ValueDict:
    def __init__(self, django_form):
        self._django_form = django_form

    def __getattribute__(self, name):
        if name == '_django_form':
            return super().__getattribute__(name)

        attribute = self._django_form._form_values.__getattribute__(name)

        if attribute not in ('clear', 'pop', 'popitem', 'update'):
            return attribute

        if not callable(attribute):
            return attribute

        def wrapper(*args, **kwargs):
            return_value = attribute(*args, **kwargs)

            self._django_form._form_data = deepcopy(
                self._django_form._form_values,
            )

            self._django_form.render()

            return return_value

        return wrapper

    def __getitem__(self, name):
        return self._django_form._form_values[name]

    def __setitem__(self, name, value):
        if not self._django_form._form_data:
            self._django_form._form_data = deepcopy(
                self._django_form._form_values,
            )

        self._django_form._form_data[name] = value

        self._django_form.render()

    def __repr__(self):
        return self._django_form._form_values.__repr__()

    def __str__(self):
        return self._django_form._form_values.__str__()

    def __len__(self):
        return self._django_form._form_values.__len__()

    def __bool__(self):
        return self._django_form._form_values.__bool__()


class DjangoForm(Widget):
    def __init__(self, form_class, *form_args, bubble_up=False,
                 rerender_on_change=True, render_as='', **form_kwargs):

        self._form_class = form_class
        self._form_args = list(deepcopy(form_args))
        self._form_kwargs = deepcopy(form_kwargs)

        self.bubble_up = bubble_up
        self.rerender_on_change = rerender_on_change
        self.render_as = render_as

        self._form_data = None
        self._initial = None

        if len(self._form_args) > 0:
            self._form_data = self._form_args.pop(0)

        if len(self._form_args) >= 4:
            self._initial = self._form_args.pop(3)

        if 'data' in self._form_kwargs:
            self._form_data = self._form_kwargs.pop('data')

        if 'initial' in self._form_kwargs:
            self._initial = self._form_kwargs.pop('initial')

        self.render()

    @property
    def values(self):
        return ValueDict(self)

    @values.setter
    def values(self, new_values):
        self._form_data = new_values

        self.render()

    def is_valid(self):
        self._form_data = deepcopy(self._form_values)
        self.render()

        return self._form.is_valid()

    def render(self):
        # render form
        self._form = self._form_class(
            *deepcopy(self._form_args),
            **{
                'data': deepcopy(self._form_data),
                'initial': deepcopy(self._initial),
                **deepcopy(self._form_kwargs),
            }
        )

        if self.rerender_on_change:
            for field in self._form.fields.values():
                field.widget.attrs['data-lona-events'] = CHANGE.serialize()

        if self.render_as:
            render_method = getattr(self._form, 'as_{}'.format(self.render_as))
            html_string = render_method()

        else:
            html_string = str(self._form)

        self.nodes = html_string_to_node_list(html_string)

        # set values
        self._form_values = {}

        for field_name, field in self._form.fields.items():
            self._form_values[field_name] = copy(field.initial)

        for field_name, value in self._form.initial.items():
            self._form_values[field_name] = copy(value)

        for field_name, value in (self._form_data or {}).items():
            self._form_values[field_name] = copy(value)

    def handle_input_event(self, input_event):
        if not self._form_data:
            self._form_data = deepcopy(self._form_values)

        field_name = input_event.node.attributes['name']
        self._form_data[field_name] = input_event.data

        if self.rerender_on_change:
            self.render()

        if self.bubble_up:
            return input_event



from copy import copy

from lona.html.nodes import TextAreaNode, InputNode, Widget
from lona.events.event_types import CHANGE

DEFAULT_INPUT_DELAY = 300


class TextInput(Widget):
    NODE_CLASS = InputNode

    ID_LIST = []
    CLASS_LIST = []
    STYLE = {}

    ATTRIBUTES = {
        'type': 'text',
    }

    def __init__(self, value=None, disabled=False, bubble_up=False,
                 input_delay=DEFAULT_INPUT_DELAY, **input_kwargs):

        self.bubble_up = bubble_up

        self.input_node = self.gen_input_node(**input_kwargs)

        self.input_node.events.add(CHANGE(input_delay))

        self.nodes = [
            self.input_node,
        ]

        self.disabled = disabled

        if value is not None:
            self.value = value

    def gen_node_args(self, **kwargs):
        id_list = copy(self.ID_LIST)
        class_list = copy(self.CLASS_LIST)
        style = copy(self.STYLE)
        attributes = copy(self.ATTRIBUTES)
        misc_kwargs = {}

        for name, value in kwargs.copy().items():

            # remove underscores from attributes
            # this makes kwargs like '_class' possible to prevent clashes
            # with python grammar
            clean_name = name

            if '_' in clean_name:
                clean_name = clean_name.replace('_', '-')

                if clean_name.startswith('-'):
                    clean_name = clean_name[1:]

            if clean_name == 'id':
                value = kwargs.pop(name)

                if isinstance(value, str):
                    value = value.split(' ')

                id_list.extend(value)

            elif clean_name == 'class':
                value = kwargs.pop(name)

                if isinstance(value, str):
                    value = value.split(' ')

                class_list.extend(value)

            elif clean_name == 'style':
                style.update(kwargs.pop(name))

            elif clean_name == 'attributes':
                attributes.update(kwargs.pop(name))

            else:
                misc_kwargs[name] = value

        return {
            'id': id_list,
            'class': class_list,
            'style': style,
            **attributes,
            **misc_kwargs,
        }

    def gen_input_node(self, **kwargs):
        return self.NODE_CLASS(**self.gen_node_args(**kwargs))

    def handle_input_event(self, input_event):
        self.value = input_event.data

        if self.bubble_up:
            return input_event

    # properties ##############################################################
    # value
    @property
    def value(self):
        return self.input_node.attributes.get('value', '')

    @value.setter
    def value(self, new_value):
        self.input_node.attributes['value'] = new_value

    # disabled
    @property
    def disabled(self):
        return self.input_nodes.attributes.get('disabled', '')

    @disabled.setter
    def disabled(self, new_value):
        with self.lock:
            if new_value:
                self.input_node.attributes['disabled'] = True

            elif 'disabled' in self.input_node.attributes:
                self.input_node.attributes.pop('disabled')

    # input node properties
    # id
    @property
    def id_list(self):
        return self.input_node.id_list

    @id_list.setter
    def id_list(self, new_value):
        self.input_node.id_list = new_value

    # class
    @property
    def class_list(self):
        return self.input_node.class_list

    @class_list.setter
    def class_list(self, new_value):
        self.input_node.class_list = new_value

    # style
    @property
    def style(self):
        return self.input_node.style

    @style.setter
    def style(self, new_value):
        self.input_node.style = new_value

    # attributes
    @property
    def attributes(self):
        return self.input_node.attributes

    @attributes.setter
    def attributes(self, new_value):
        self.input_node.attributes = new_value

    # ignore
    @property
    def ignore(self):
        return self.input_node.ignore

    @ignore.setter
    def ignore(self, new_value):
        self.input_node.ignore = new_value

    # events
    @property
    def events(self):
        return self.input_node.events

    @events.setter
    def events(self, new_value):
        self.input_node.events = new_value

    # helper
    @property
    def has_id(self):
        return self.input_node.has_id

    @property
    def has_class(self):
        return self.input_node.has_class


class TextArea(TextInput):
    NODE_CLASS = TextAreaNode


class CheckBox(TextInput):
    ATTRIBUTES = {
        'type': 'checkbox',
    }

    @property
    def value(self):
        return self.input_node.attributes.get('checked', False)

    @value.setter
    def value(self, new_value):
        self.input_node.attributes['checked'] = bool(new_value)



class Symbol:
    _INCLUDE_IN_FRONTEND_LIBRARY = False

    def __init__(self, name, value=None):
        if not isinstance(name, str):
            raise ValueError('name has to be a string')

        self.name = name.strip().upper()

        if value is not None:
            self.value = value

        else:
            self.value = self.name

    def __repr__(self):
        return self.name

    @classmethod
    def dump_symbol_classes(cls):
        data = {}

        for symbol_class in cls.__subclasses__():
            if not symbol_class._INCLUDE_IN_FRONTEND_LIBRARY:
                continue

            class_name = symbol_class.__name__
            data[class_name] = {}

            for name in dir(symbol_class):
                if name.startswith('_'):
                    continue

                attribute = getattr(symbol_class, name)

                if not isinstance(attribute, Symbol):
                    continue

                data[class_name][attribute.name] = attribute.value

        return data

    # comparisons #############################################################
    def __eq__(self, other):
        other_value = other

        if isinstance(other_value, Symbol):
            other_value = other_value.value

        if isinstance(other_value, str):
            other_value = other_value.strip().upper()

        return (other_value == self.name or
                other_value == self.value)

    def _compare(self, other, operator):
        if isinstance(other, Symbol):
            return operator(self.value, other.value)

        return operator(self.value, other)

    def __lt__(self, other):
        return self._compare(other, lambda a, b: a < b)

    def __lte__(self, other):
        return self._compare(other, lambda a, b: a <= b)

    def __gt__(self, other):
        return self._compare(other, lambda a, b: a > b)

    def __gte__(self, other):
        return self._compare(other, lambda a, b: a >= b)


class Mapping:
    def __init__(self, entries=None):
        self.entries = entries or []

    def get_entry(self, key):
        for entry in self.entries:
            if entry[0] == key:
                return entry

    def keys(self):
        return [i[0] for i in self.entries]

    def items(self):
        for entry in self.entries:
            yield entry

    def pop(self, key):
        for entry in self.entries:
            if entry[0] == key:
                self.entries.remove(entry)

                return entry[1]

        raise KeyError(key)

    def copy(self):
        return Mapping(self.entries.copy())

    def __getitem__(self, key):
        entry = self.get_entry(key)

        if not entry:
            raise KeyError(key)

        return entry[1]

    def __setitem__(self, key, value):
        entry = self.get_entry(key)

        if entry:
            entry[1] = value

        else:
            self.entries.append([key, value])

    def __contains__(self, key):
        return bool(self.get_entry(key))

    def __repr__(self):
        d = {}

        for key, value in self.entries:
            d[repr(key)] = value

        return repr(d)



from concurrent.futures import CancelledError
from datetime import datetime
import threading
import logging
import asyncio
import time

from yarl import URL

from lona.exceptions import StopReason, ServerStop, UserAbort
from lona.html.abstract_node import AbstractNode
from lona.events.input_event import InputEvent
from lona.symbols import VIEW_RUNTIME_STATE
from lona.html.document import Document
from lona.errors import ForbiddenError
from lona.request import Request

from lona.protocol import (
    encode_http_redirect,
    encode_view_start,
    encode_view_stop,
    encode_redirect,
    encode_data,
)


logger = logging.getLogger('lona.view_runtime')


class ViewRuntime:
    def __init__(self, server, url, route, match_info, post_data={},
                 frontend=False, start_connection=None):

        self.server = server
        self.url = URL(url or '')
        self.route = route
        self.match_info = match_info
        self.post_data = post_data
        self.frontend = frontend
        self.start_connection = start_connection

        # find view
        if route:
            self.view = route.view

        else:
            self.view = self.server.settings.CORE_ERROR_404_VIEW

        if frontend:
            self.view = self.server.settings.CORE_FRONTEND_VIEW

            if route and route.frontend_view:
                self.view = route.frontend_view

        self.view_class = self.server.view_loader.load(self.view)
        self.name = repr(self.view_class)
        self.view = self.view_class()

        # setup state
        self.connections = {}
        # contains: {
        #     connection: (window_id, url),
        # }

        self.document = Document()

        self.stopped = asyncio.Future(loop=self.server.loop)
        self.is_stopped = False
        self.stop_reason = None
        self.is_daemon = False

        self.view_runtime_id = time.monotonic_ns()
        self.state = VIEW_RUNTIME_STATE.NOT_STARTED
        self.thread_ident = None
        self.started_at = None
        self.stopped_at = None

        self.pending_input_events = {
            'event': None,
            'click': None,
            'change': None,
            'submit': None,
        }

        self.request = Request(
            view_runtime=self,
            connection=self.start_connection,
        )

        self.view_args = (self.request,)
        self.view_kwargs = {}

        if not self.frontend:
            self.view_kwargs = dict(self.match_info or {})

    # state ###################################################################
    @property
    def state(self):
        if self.is_stopped:
            if self.stop_reason is not None:
                if isinstance(self.stop_reason, UserAbort):
                    return VIEW_RUNTIME_STATE.STOPPED_BY_USER

                elif isinstance(self.stop_reason, ServerStop):
                    return VIEW_RUNTIME_STATE.STOPPED_BY_SERVER

                else:
                    return VIEW_RUNTIME_STATE.CRASHED

            else:
                return VIEW_RUNTIME_STATE.FINISHED

        return self._state

    @state.setter
    def state(self, new_state):
        self._state = new_state

    # middlewares #############################################################
    def run_middlewares(self):
        try:
            handled, raw_response_dict, middleware = \
                self.server.middleware_controller.process_request(
                    view=self.view,
                    request=self.request,
                )

            if handled:
                if not raw_response_dict:
                    raw_response_dict = ''

                self.send_view_start()

                response_dict = self.handle_raw_response_dict(
                    raw_response_dict,
                )

                self.send_view_stop()

                return response_dict

        except Exception as exception:
            logger.error(
                'Exception raised while running middleware hooks',
                exc_info=True,
            )

            self.send_view_start()

            view_class = self.server.view_loader.load(
                self.server.settings.CORE_ERROR_500_VIEW,
            )

            view = view_class()

            return self.handle_raw_response_dict(
                view.handle_request(
                    self.request,
                    exception=exception,
                )
            )

    # permission checks #######################################################
    def run_user_enter(self):
        try:
            try:
                self.view.handle_user_enter(self.request)

            except ForbiddenError as exception:
                self.send_view_start()

                view_class = self.server.view_loader.load(
                    self.server.settings.CORE_ERROR_403_VIEW,
                )

                view = view_class()

                response_dict = self.handle_raw_response_dict(
                    view.handle_request(
                        self.request,
                        exception=exception,
                    ),
                )

                self.send_view_stop()

                return response_dict

        except Exception as exception:
            logger.error(
                'Exception raised while running handle_user_enter',
                exc_info=True,
            )

            self.send_view_start()

            view_class = self.server.view_loader.load(
                self.server.settings.ERROR_500_VIEW,
            )

            view = view_class()

            return self.handle_raw_response_dict(
                view.handle_request(
                    self.request,
                    exception=exception,
                )
            )

    # start and stop ##########################################################
    def start(self):
        try:
            # update internal state
            self.thread_ident = threading.current_thread().ident
            self.state = VIEW_RUNTIME_STATE.RUNNING
            self.started_at = datetime.now()

            # start view
            self.send_view_start()

            raw_response_dict = self.view.handle_request(
                *self.view_args,
                **self.view_kwargs,
            )

            # response dicts
            if raw_response_dict:
                if isinstance(raw_response_dict, AbstractNode):
                    self.request.client.show(html=raw_response_dict)

                    return

                # check if non-interactive features got used in
                # interactive mode
                if(self.route and self.route.interactive and
                    isinstance(raw_response_dict, dict) and (
                     'json' in raw_response_dict or
                     'file' in raw_response_dict)):

                    raise RuntimeError('JSON and file responses are only available in non-interactive mode')  # NOQA

                return self.handle_raw_response_dict(raw_response_dict)

        except(StopReason, CancelledError):
            pass

        # 403 Forbidden
        except ForbiddenError as exception:
            view_class = self.server.view_loader.load(
                self.server.settings.CORE_ERROR_403_VIEW,
            )

            view = view_class()

            return self.handle_raw_response_dict(
                view.handle_request(
                    self.request,
                    exception=exception,
                )
            )

        # 500 Internal Error
        except Exception as exception:
            self.stop_reason = exception

            logger.error(
                'Exception raised while running %s',
                self.view,
                exc_info=True,
            )

            view_class = self.server.view_loader.load(
                self.server.settings.CORE_ERROR_500_VIEW,
            )

            view = view_class()

            return self.handle_raw_response_dict(
                view.handle_request(
                    self.request,
                    exception=exception,
                )
            )

        finally:
            self.is_stopped = True
            self.stopped_at = datetime.now()
            self.send_view_stop()

    def stop(self, reason=UserAbort, clean_up=True):
        self.stop_reason = reason

        if not isinstance(self.stop_reason, Exception):
            self.stop_reason = self.stop_reason()

        # let all calls on request.view.await_sync() crash with
        # the set stop reason
        def _set_stopped():
            if not self.stopped.done():
                self.stopped.set_result(True)

        self.server.loop.call_soon_threadsafe(_set_stopped)

        # cancel all pending user input events
        for pending_input_event in self.pending_input_events.copy().items():
            if not pending_input_event[1]:
                continue

            name, (future, nodes) = pending_input_event

            if not future.done():
                future.set_exception(self.stop_reason)

        # clean up
        if clean_up:
            # drop all connections
            self.connections = {}

            # multi user views have no start_connection
            if self.start_connection:
                self.server.view_runtime_controller.remove_view_runtime(self)

    def issue_500_error(self, exception):
        # stop the runtime but don't run cleanup code to get the
        # output of the 500 handle through
        self.stop(reason=exception, clean_up=False)

        # run 500 view
        # this runs if the crash happened in an input event handler after
        # the view stopped
        if self.is_stopped:
            view_class = self.server.view_loader.load(
                self.server.settings.CORE_ERROR_500_VIEW,
            )

            view = view_class()

            self.handle_raw_response_dict(
                view.handle_request(
                    self.request,
                    exception=exception,
                )
            )

            logger.error(
                'Exception raised after running %s',
                self.view,
                exc_info=True,
            )

    # connection management ###################################################
    def add_connection(self, connection, window_id, url):
        self.connections[connection] = (window_id, url, )

        with self.document.lock:
            data_type, data = self.document.serialize()

            if not data:
                return

            self.send_data(
                data=[data_type, data],
                connections={connection: (window_id, url, )},
            )

    def reconnect_connection(self, connection, window_id, url):
        self.connections[connection] = (window_id, url, )

        with self.document.lock:
            self.send_view_start()

            data_type, data = self.document.serialize()

            if not data:
                return

            self.send_data(
                data=[data_type, data],
                connections={connection: (window_id, url, )},
            )

    def remove_connection(self, connection, window_id=None):
        if connection in self.connections:
            if window_id:
                if self.connections[connection] == window_id:
                    self.connections.pop(connection)

            else:
                self.connections.pop(connection)

        # if the last connection gets closed and the view should
        # not continue running in background, it gets stopped
        if not self.connections and not self.is_daemon:
            self.stop()

    # lona messages ###########################################################
    def send_redirect(self, target_url, connections={}):
        connections = connections or self.connections

        for connection, (window_id, url) in connections.items():
            message = encode_redirect(
                window_id=window_id,
                view_runtime_id=self.view_runtime_id,
                target_url=target_url,
            )

            connection.send_str(message)

    def send_http_redirect(self, target_url, connections={}):
        connections = connections or self.connections

        for connection, (window_id, url) in connections.items():
            message = encode_http_redirect(
                window_id=window_id,
                view_runtime_id=self.view_runtime_id,
                target_url=target_url,
            )

            connection.send_str(message)

    def send_data(self, title=None, data=None, connections={}):
        connections = connections or self.connections

        # send message
        for connection, (window_id, url) in list(connections.items()):
            message = encode_data(
                window_id=window_id,
                view_runtime_id=self.view_runtime_id,
                title=title,
                data=data,
            )

            connection.send_str(message)

    def send_view_start(self, connections={}):
        connections = connections or self.connections

        for connection, (window_id, url) in list(connections.items()):
            message = encode_view_start(
                window_id=window_id,
                view_runtime_id=self.view_runtime_id,
            )

            connection.send_str(message)

    def send_view_stop(self, connections={}):
        connections = connections or self.connections

        for connection, (window_id, url) in list(connections.items()):
            message = encode_view_stop(
                window_id=window_id,
                view_runtime_id=self.view_runtime_id,
            )

            connection.send_str(message)

    def handle_raw_response_dict(self, raw_response_dict, connections={}):
        connections = connections or self.connections

        response_dict = self.server.response_parser.render_response_dict(
            raw_response_dict=raw_response_dict,
            view_name=self.name,
        )

        if response_dict['redirect']:
            self.send_redirect(
                response_dict['redirect'],
                connections=connections,
            )

        elif response_dict['http_redirect']:
            self.send_http_redirect(
                response_dict['http_redirect'],
                connections=connections,
            )

        elif response_dict['text']:
            with self.document.lock:
                data = self.document.apply(response_dict['text'])
                self.send_data(data=data, connections=connections)

        return response_dict

    # input events ############################################################
    def await_input_event(self, nodes=[], event_type='event'):
        async def _await_input_event():
            future = asyncio.Future()
            self.pending_input_events[event_type] = [future, nodes]

            return await future

        return self.server.run_coroutine_sync(_await_input_event())

    def handle_input_event(self, connection, payload):
        request = Request(
            view_runtime=self,
            connection=connection,
        )

        input_event = InputEvent(
            request=request,
            payload=payload,
            document=self.document,
        )

        def send_html_update():
            with self.document.lock:
                data = self.document.apply(self.document.html)

                if data:
                    self.send_data(data=data)

        try:
            # input event root handler
            input_event = self.view.handle_input_event_root(input_event)

            if not input_event:
                send_html_update()

                return

            # widgets
            for widget in input_event.widgets[::-1]:
                input_event = widget.handle_input_event(input_event)

                if not input_event:
                    send_html_update()

                    return

            # pending input events
            if(input_event.name in self.pending_input_events and
               self.pending_input_events[input_event.name] is not None):

                future, nodes = self.pending_input_events[input_event.name]

                if not nodes or input_event.node in nodes:
                    future.set_result(input_event)
                    self.pending_input_events[input_event.name] = None

                    return

            if self.pending_input_events['event'] is not None:
                future, nodes = self.pending_input_events['event']

                if not nodes or input_event.node in nodes:
                    future.set_result(input_event)
                    self.pending_input_events['event'] = None

                    return

            # input event handler (class based views)
            input_event = self.view.handle_input_event(input_event)

            if not input_event:
                send_html_update()

        except(StopReason, CancelledError):
            pass

        except Exception as exception:
            self.issue_500_error(exception)



from time import monotonic_ns
from copy import deepcopy

from lona.protocol import OPERATION, PATCH_TYPE


def check_value(value):
    if value is None:
        return

    if not isinstance(value, (bool, int, float, str, list, dict)):
        raise ValueError('unsupported type: {}'.format(type(value)))

    if isinstance(value, list):
        for i in value:
            check_value(i)

    if isinstance(value, dict):
        for k, v in value.items():
            check_value(k)
            check_value(v)


class ListOverlay:
    def __init__(self, widget_data, key_path, original_data):
        self._widget_data = widget_data
        self._widget = widget_data._widget
        self._key_path = key_path
        self._original_data = original_data

    def append(self, item):
        check_value(item)

        with self._widget.lock:
            self._original_data.append(item)

            self._widget_data._changes.append([
                monotonic_ns(),
                self._widget._id,
                PATCH_TYPE.WIDGET_DATA,
                self._key_path,
                OPERATION.INSERT,
                len(self._original_data) - 1,
                deepcopy(item),
            ])

    def clear(self):
        with self._widget.lock:
            self._original_data.clear()

            self._widget_data._changes.append([
                monotonic_ns(),
                self._widget._id,
                PATCH_TYPE.WIDGET_DATA,
                self._key_path,
                OPERATION.CLEAR,
            ])

    def copy(self, *args, **kwargs):
        with self._widget.lock:
            return self._original_data.copy(*args, **kwargs)

    def count(self, *args, **kwargs):
        with self._widget.lock:
            return self._original_data.count(*args, **kwargs)

    def extend(self, items):
        check_value(items)

        with self._widget.lock:
            for item in items:
                self.append(item)

    def index(self, *args, **kwargs):
        with self._widget.lock:
            return self._original_data.count(*args, **kwargs)

    def insert(self, index, item):
        check_value(item)

        with self._widget.lock:
            self._original_data.insert(index, item)

            self._widget_data._changes.append([
                monotonic_ns(),
                self._widget._id,
                PATCH_TYPE.WIDGET_DATA,
                self._key_path,
                OPERATION.INSERT,
                self._original_data.index(item),
                deepcopy(item),
            ])

    def pop(self, index):
        with self._widget.lock:
            item = self._original_data.pop(index)

            self._widget_data._changes.append([
                monotonic_ns(),
                self._widget._id,
                PATCH_TYPE.WIDGET_DATA,
                self._key_path,
                OPERATION.REMOVE,
                index,
            ])

            return item

    def remove(self, item):
        with self._widget.lock:
            index = 0

            for i in self._original_data:
                if i == item:
                    self._widget_data._changes.append([
                        monotonic_ns(),
                        self._widget._id,
                        PATCH_TYPE.WIDGET_DATA,
                        self._key_path,
                        OPERATION.REMOVE,
                        index,
                    ])

                index += 1

            self._original_data.remove(item)

    def reverse(self, *args, **kwargs):
        raise NotImplementedError

    def sort(self, *args, **kwargs):
        raise NotImplementedError

    def __setitem__(self, name, item):
        with self._widget.lock:
            check_value(name)
            check_value(item)

            self._original_data[name] = item

            self._widget_data._changes.append([
                monotonic_ns(),
                self._widget._id,
                PATCH_TYPE.WIDGET_DATA,
                self._key_path,
                OPERATION.SET,
                name,
                item,
            ])

    def __getitem__(self, name):
        with self._widget.lock:
            item = self._original_data[name]

            if isinstance(item, list):
                return ListOverlay(
                    widget_data=self._widget_data,
                    key_path=self._key_path + [name],
                    original_data=item,
                )

            if isinstance(item, dict):
                return DictOverlay(
                    widget_data=self._widget_data,
                    key_path=self._key_path + [name],
                    original_data=item,
                )

            return item

    def __delitem__(self, name):
        with self._widget.lock:
            del self._original_data[name]

            self._widget_data._changes.append([
                monotonic_ns(),
                self._widget._id,
                PATCH_TYPE.WIDGET_DATA,
                self._key_path,
                OPERATION.REMOVE,
                name,
            ])

    def __str__(self, *args, **kwargs):
        with self._widget.lock:
            return str(self._original_data)

    def __repr__(self, *args, **kwargs):
        with self._widget.lock:
            return repr(self._original_data)

    def __len__(self, *args, **kwargs):
        with self._widget.lock:
            return len(self._original_data)

    def __bool__(self, *args, **kwargs):
        with self._widget.lock:
            return bool(self._original_data)


class DictOverlay:
    def __init__(self, widget_data, key_path, original_data):
        self._widget_data = widget_data
        self._widget = widget_data._widget
        self._key_path = key_path
        self._original_data = original_data

    def clear(self):
        with self._widget.lock:
            self._original_data.clear()

            self._widget_data._changes.append([
                monotonic_ns(),
                self._widget._id,
                PATCH_TYPE.WIDGET_DATA,
                self._key_path,
                OPERATION.CLEAR,
            ])

    def copy(self, *args, **kwargs):
        with self._widget.lock:
            return self._original_data.copy(*args, **kwargs)

    def fromkeys(self, *args, **kwargs):
        raise NotImplementedError

    def get(self, *args, **kwargs):
        with self._widget.lock:
            return self._original_data.get(*args, **kwargs)

    def items(self, *args, **kwargs):
        with self._widget.lock:
            return self._original_data.items()

    def keys(self, *args, **kwargs):
        with self._widget.lock:
            return self._original_data.keys()

    def pop(self, key):
        with self._widget.lock:
            item = self._original_data.pop(key)

            self._widget_data._changes.append([
                monotonic_ns(),
                self._widget._id,
                PATCH_TYPE.WIDGET_DATA,
                self._key_path,
                OPERATION.REMOVE,
                key,
            ])

            return item

    def popitem(self):
        with self._widget.lock:
            key, value = self._original_data.popitem()

            self._widget_data._changes.append([
                monotonic_ns(),
                self._widget._id,
                PATCH_TYPE.WIDGET_DATA,
                self._key_path,
                OPERATION.REMOVE,
                key,
            ])

            return key, value

    def setdefault(self, *args, **kwargs):
        raise NotImplementedError

    def update(self, update_dict):
        with self._widget.lock:
            for key, value in update_dict.items():
                self._original_data[key] = value

                self._widget_data._changes.append([
                    monotonic_ns(),
                    self._widget._id,
                    PATCH_TYPE.WIDGET_DATA,
                    self._key_path,
                    OPERATION.SET,
                    key,
                    value,
                ])

    def values(self, *args, **kwargs):
        with self._widget.lock:
            return self._original_data.values()

    def __setitem__(self, name, item):
        with self._widget.lock:
            check_value(name)
            check_value(item)

            self._original_data[name] = item

            self._widget_data._changes.append([
                monotonic_ns(),
                self._widget._id,
                PATCH_TYPE.WIDGET_DATA,
                self._key_path,
                OPERATION.SET,
                name,
                item,
            ])

    def __getitem__(self, name):
        with self._widget.lock:
            item = self._original_data[name]

            if isinstance(item, list):
                return ListOverlay(
                    widget_data=self._widget_data,
                    key_path=self._key_path + [name],
                    original_data=item,
                )

            if isinstance(item, dict):
                return DictOverlay(
                    widget_data=self._widget_data,
                    key_path=self._key_path + [name],
                    original_data=item,
                )

            return item

    def __delitem__(self, name):
        with self._widget.lock:
            del self._original_data[name]

            self._widget_data._changes.append([
                monotonic_ns(),
                self._widget._id,
                PATCH_TYPE.WIDGET_DATA,
                self._key_path,
                OPERATION.REMOVE,
                name,
            ])

    def __str__(self, *args, **kwargs):
        with self._widget.lock:
            return str(self._original_data)

    def __repr__(self, *args, **kwargs):
        with self._widget.lock:
            return repr(self._original_data)

    def __len__(self, *args, **kwargs):
        with self._widget.lock:
            return len(self._original_data)

    def __bool__(self, *args, **kwargs):
        with self._widget.lock:
            return bool(self._original_data)


class WidgetData:
    _LONA_CLASS_NAME = 'WidgetData'

    def __init__(self, widget):
        self._widget = widget
        self._changes = []

        self._reset({})

    def __getitem__(self, *args, **kwargs):
        return self._overlay.__getitem__(*args, **kwargs)

    def __setitem__(self, *args, **kwargs):
        return self._overlay.__setitem__(*args, **kwargs)

    def __delitem__(self, *args, **kwargs):
        return self._overlay.__delitem__(*args, **kwargs)

    def __len__(self, *args, **kwargs):
        with self._widget.lock:
            return len(self._original_data)

    def __bool__(self, *args, **kwargs):
        with self._widget.lock:
            return bool(self._original_data)

    def __getattribute__(self, name):
        if name.startswith('_'):
            return super().__getattribute__(name)

        if hasattr(self._overlay, name):
            return getattr(self._overlay, name)

        return super().__getattribute__(name)

    def __dir__(self):
        return [
            *super().__dir__(),
            *dir(self._overlay),
        ]

    # serialisation ###########################################################
    def _reset(self, value):
        if not isinstance(value, (dict, list)):
            raise ValueError('widget state has to be dict or list')

        check_value(value)

        with self._widget.lock:
            self._data = value

            self._changes.append([
                monotonic_ns(),
                self._widget._id,
                PATCH_TYPE.WIDGET_DATA,
                [],
                OPERATION.RESET,
                deepcopy(value)
            ])

            if isinstance(value, list):
                self._overlay = ListOverlay(
                    widget_data=self,
                    key_path=[],
                    original_data=self._data
                )

            elif isinstance(value, dict):
                self._overlay = DictOverlay(
                    widget_data=self,
                    key_path=[],
                    original_data=self._data
                )

    def _has_changes(self):
        return bool(self._changes)

    def _get_changes(self):
        return self._changes.copy()

    def _clear_changes(self):
        self._changes = []

    def _serialize(self):
        return deepcopy(self._data)

    # string representation ###################################################
    def __repr__(self):
        return '<WidgetData({}))>'.format(repr(self._data))



import logging

from lona.protocol import PROTOCOL, EXIT_CODE, decode_message

logger = logging.getLogger('lona.protocol')


class AnonymousUser:
    def __repr__(self):
        return '<AnonymousUser()>'

    def __eq__(self, other):
        return isinstance(other, AnonymousUser)


class LonaMessageMiddleware:
    def process_connection(self, data):
        if data.connection.user is None:
            data.connection.user = AnonymousUser()

        return data

    def process_websocket_message(self, data):
        if not data.message.startswith(PROTOCOL.MESSAGE_PREFIX.value):
            return data

        exit_code, window_id, view_runtime_id, method, payload = \
            decode_message(data.message)

        if exit_code != EXIT_CODE.SUCCESS:
            logger.error('invalid lona message received: %s', data.message)

            return data

        data.server.view_runtime_controller.handle_lona_message(
            connection=data.connection,
            window_id=window_id,
            view_runtime_id=view_runtime_id,
            method=method,
            payload=payload,
        )



from lona.html.data_binding.select import *  # NOQA
from lona.html.data_binding.inputs import *  # NOQA
from lona.html.widget import Widget  # NOQA
from lona.html.widgets import HTML  # NOQA
from lona.html.nodes import *  # NOQA



import logging

from django.contrib.auth.models import User, AnonymousUser
from django.contrib.sessions.models import Session

logger = logging.getLogger('lona.contrib.django.auth')


class DjangoSessionMiddleware:
    def process_connection(self, data):
        # find user
        logger.debug('searching for django user')

        if isinstance(data.connection, User):
            return data

        session_key = data.connection.http_request.cookies.get('sessionid', '')
        user = AnonymousUser()

        if session_key:
            try:
                session = Session.objects.get(session_key=session_key)
                uid = session.get_decoded().get('_auth_user_id')

                try:
                    user = User.objects.get(pk=uid)

                except User.DoesNotExist:
                    pass

            except Session.DoesNotExist:
                pass

        logger.debug('user set to %s', user)

        data.connection.user = user

        return data



from copy import copy

from lona.html.nodes import SelectNode, OptionNode, Widget
from lona.events.event_types import CHANGE


class Select(Widget):
    ID_LIST = []
    CLASS_LIST = []
    STYLE = {}
    ATTRIBUTES = {}

    def __init__(self, values=[], disabled=False, bubble_up=False,
                 **select_kwargs):

        self.bubble_up = bubble_up

        self.select_node = SelectNode(**self.gen_node_args(**select_kwargs))
        self.select_node.events.add(CHANGE)

        self.nodes = [
            self.select_node,
        ]

        self.values = values

    def gen_node_args(self, **kwargs):
        id_list = copy(self.ID_LIST)
        class_list = copy(self.CLASS_LIST)
        style = copy(self.STYLE)
        attributes = copy(self.ATTRIBUTES)

        if '_id' in kwargs:
            id_list.extend(kwargs.pop('_id'))

        if '_class' in kwargs:
            id_list.extend(kwargs.pop('_class'))

        if 'style' in kwargs:
            style.update(kwargs.pop('style'))

        elif '_style' in kwargs:
            style.update(kwargs.pop('_style'))

        if 'attributes' in kwargs:
            attributes.update(kwargs.pop('attributes'))

        elif '_attributes' in kwargs:
            attributes.update(kwargs.pop('_attributes'))

        return {
            'id': id_list,
            'class': class_list,
            'style': style,
            **attributes,
            **kwargs,
        }

    def handle_input_event(self, input_event):
        with self.lock:
            self.value = input_event.data

            if self.bubble_up:
                return input_event

    @property
    def disabled(self):
        return self.input_nodes.attributes.get('disabled', '')

    @disabled.setter
    def disabled(self, new_value):
        with self.lock:
            if new_value:
                self.input_node.attributes['disabled'] = True

            elif 'disabled' in self.input_node.attributes:
                self.input_node.attributes.pop('disabled')

    @property
    def values(self):
        with self.lock:
            return self._values

    @values.setter
    def values(self, new_values):
        with self.lock:
            self._values = []
            self.select_node.clear()

            for i in new_values:
                value, name, selected = (list(i) + [False])[0:3]
                option_node = OptionNode(str(name), value=str(value))

                if selected:
                    option_node.attributes['selected'] = ''

                self.select_node.append(option_node)

                self._values.append(
                    (value, name, ),
                )

    @property
    def value(self):
        with self.lock:
            value = []

            for option in self.select_node.nodes:
                if 'selected' in option.attributes:
                    value.append(option.attributes['value'])

            if not value:
                option = self.select_node.nodes[0]

                value.append(option.attributes['value'])

            if 'multiple' not in self.select_node.attributes:
                value = value.pop()

            return value

    @value.setter
    def value(self, new_value):
        if not isinstance(new_value, list):
            new_value = [new_value]

        with self.lock:
            for option in self.select_node.nodes:
                if option.attributes['value'] in new_value:
                    if 'selected' not in option.attributes:
                        option.attributes['selected'] = ''

                else:
                    if 'selected' in option.attributes:
                        option.attributes.pop('selected')



from time import monotonic_ns

from lona.html.abstract_node import AbstractNode
from lona.protocol import OPERATION, PATCH_TYPE
from lona.html.text_node import TextNode


class NodeList:
    def __init__(self, node):
        self._node = node
        self._nodes = []
        self._changes = []

    # list helper #############################################################
    def _check_node(self, node):
        if isinstance(node, (str, int, float, bool, )):
            node = TextNode(node)

        if not isinstance(node, AbstractNode):
            raise ValueError('unsupported type: {}'.format(repr(node)))

        return node

    def _prepare_node(self, node):
        if node.parent and node.parent is not self._node:
            node.parent.remove(node)

        elif node == self._node.root:
            raise RuntimeError('loop detected')

        node.parent = self._node
        node._clear_changes()

    def insert(self, index, node):
        node = self._check_node(node)

        with self._node.document.lock:
            self._prepare_node(node)
            self._nodes.insert(index, node)

            index = self._nodes.index(node)

            self._changes.append([
                monotonic_ns(),
                self._node._id,
                PATCH_TYPE.NODES,
                OPERATION.INSERT,
                index,
                node._serialize(),
            ])

    def append(self, node):
        node = self._check_node(node)

        with self._node.document.lock:
            self._prepare_node(node)
            self._nodes.append(node)

            index = self._nodes.index(node)

            self._changes.append([
                monotonic_ns(),
                self._node._id,
                PATCH_TYPE.NODES,
                OPERATION.INSERT,
                index,
                node._serialize(),
            ])

    def remove(self, node):
        with self._node.document.lock:
            self._nodes.remove(node)

            node.parent = None

            self._changes.append([
                monotonic_ns(),
                self._node._id,
                PATCH_TYPE.NODES,
                OPERATION.REMOVE,
                node._id,
            ])

    def pop(self, index):
        with self._node.document.lock:
            node = self._nodes.pop(index)

            node.parent = None

            self._changes.append([
                monotonic_ns(),
                self._node._id,
                PATCH_TYPE.NODES,
                OPERATION.REMOVE,
                node._id,
            ])

            return node

    def clear(self):
        with self._node.document.lock:
            if not self._nodes:
                return

            for node in list(self._nodes):
                node.parent = None
                self._nodes.remove(node)

            self._changes.append([
                monotonic_ns(),
                self._node._id,
                PATCH_TYPE.NODES,
                OPERATION.CLEAR,
            ])

    def __getitem__(self, index):
        with self._node.document.lock:
            return self._nodes[index]

    def __setitem__(self, index, node):
        node = self._check_node(node)

        with self._node.document.lock:
            self._prepare_node(node)
            self._nodes[index] = node

            self._changes.append([
                monotonic_ns(),
                self._node._id,
                PATCH_TYPE.NODES,
                OPERATION.SET,
                index,
                node._serialize(),
            ])

    def __bool__(self):
        with self._node.document.lock:
            return bool(self._nodes)

    def __len__(self):
        with self._node.document.lock:
            return self._nodes.__len__()

    def __iter__(self):
        with self._node.document.lock:
            return self._nodes.__iter__()

    # serialisation ###########################################################
    def _reset(self, value):
        if not isinstance(value, list):
            value = [value]

        with self._node.document.lock:
            self._nodes.clear()

            for node in value:
                node = self._check_node(node)
                self._prepare_node(node)
                self._nodes.append(node)

                self._changes.append([
                    monotonic_ns(),
                    self._node._id,
                    PATCH_TYPE.NODES,
                    OPERATION.RESET,
                    [i._serialize() for i in self._nodes]
                ])

    def _has_changes(self):
        return bool(self._changes)

    def _get_changes(self):
        return list(self._changes)

    def _clear_changes(self):
        for node in self._nodes:
            node._clear_changes()

        self._changes.clear()

    def _serialize(self):
        return [i._serialize() for i in self._nodes]

    # string representation ###################################################
    def __str__(self):
        with self._node.document.lock:
            return '\n'.join([str(i) for i in self._nodes])

    def __repr__(self):
        return '<NodeList({}))>'.format(repr(self._nodes))



import builtins
import logging

from jinja2 import Environment, FileSystemLoader

logger = logging.getLogger('lona.templating')

BUILTINS = {}

for name in dir(builtins):
    if name.startswith('_'):
        continue

    BUILTINS[name] = getattr(builtins, name)


class Namespace:
    def __init__(self, server):
        self.server = server

    def load_stylesheets(self):
        return self.server.static_file_loader.style_tags_html

    def load_scripts(self):
        return self.server.static_file_loader.script_tags_html

    def _import(self, *args, **kwargs):
        return self.server.acquire(*args, **kwargs)

    def resolve_url(self, *args, **kwargs):
        return self.server.router.reverse(*args, **kwargs)

    def __getattribute__(self, name):
        # this is necessary because in python its illegal to name a
        # function 'import'

        if name == 'import':
            return super().__getattribute__('_import')

        return super().__getattribute__(name)


class TemplatingEngine:
    def __init__(self, server):
        self.server = server

        self.template_dirs = (self.server.settings.TEMPLATE_DIRS +
                              self.server.settings.CORE_TEMPLATE_DIRS)

        # resolving potential relative paths
        for index, template_dir in enumerate(self.template_dirs):
            self.template_dirs[index] = self.server.resolve_path(template_dir)

        logger.debug('loading template_dirs %s',
                     repr(self.template_dirs)[1:-1])

        self.jinja2_env = Environment(
            loader=FileSystemLoader(self.template_dirs),
        )

    # public api ##############################################################
    def get_template(self, template_name):
        logger.debug("searching for '%s'", template_name)

        template = self.jinja2_env.get_template(template_name)

        logger.debug("'%s' is '%s' ", template_name, template.filename)

        return template

    def generate_template_context(self, overrides):
        context = {
            'Lona': Namespace(self.server),
            **BUILTINS,
            **self.server.settings.TEMPLATE_EXTRA_CONTEXT,
            **overrides,
        }

        context['_variables'] = context

        return context

    def render_string(self, template_string, template_context={}):
        template = self.jinja2_env.from_string(template_string)

        template_context = self.generate_template_context(
            overrides=template_context)

        return template.render(template_context)

    def render_template(self, template_name, template_context={}):
        template = self.get_template(template_name)

        template_context = self.generate_template_context(
            overrides=template_context)

        return template.render(template_context)



from lona.types import Symbol


class VIEW_RUNTIME_STATE(Symbol):
    NOT_STARTED = Symbol('NOT_STARTED', 10)

    RUNNING = Symbol('RUNNING', 21)
    WAITING_FOR_IOLOOP = Symbol('WAITING_FOR_IOLOOP', 22)
    SLEEPING = Symbol('SLEEPING', 23)
    WAITING_FOR_INPUT = Symbol('WAITING_FOR_INPUT', 24)

    FINISHED = Symbol('FINISHED', 31)
    CRASHED = Symbol('CRASHED', 32)
    STOPPED_BY_USER = Symbol('STOPPED_BY_USER', 33)
    STOPPED_BY_SERVER = Symbol('STOPPED_BY_SERVER', 34)



from traceback import format_exception
from textwrap import indent
import threading
import datetime
import logging
import socket


class LogFilter(logging.Filter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.excluded = []
        self.included = []

    def include(self, logger_name):
        self.included.append(logger_name)

    def exclude(self, logger_name):
        self.excluded.append(logger_name)

    def filter(self, record):
        # filter exceptions that lona.command_line.run_server.run_server
        # handles it self
        if record.exc_info:

            # OSErrors
            if(isinstance(record.exc_info[1], OSError) and
               record.exc_info[1].errno in (13, 98)):

                return False

            # socket.gaierror
            if(isinstance(record.exc_info[1], socket.gaierror) and
               record.exc_info[1].errno in (-2, )):

                return False

        if record.name in self.excluded:
            return False

        if self.included and record.name not in self.included:
            return False

        return True


class LogFormatter(logging.Formatter):
    def format(self, record):
        current_thread_name = threading.current_thread().name

        # format record string
        time_stamp = datetime.datetime.fromtimestamp(record.created)
        time_stamp_str = time_stamp.strftime('%H:%M:%S.%f')

        record_string = '{}{} {}{} {} {} {}'.format(
            current_thread_name,
            (30 - len(current_thread_name)) * ' ',

            record.levelname,
            (8 - len(record.levelname)) * ' ',

            time_stamp_str,
            record.name,
            record.getMessage(),
        )

        # format exc_info
        if record.exc_info:
            record_string = '{}\n{}'.format(
                record_string,
                indent(
                    ''.join(format_exception(*record.exc_info))[:-1],
                    prefix='  ',
                ),
            )

        # colors
        RED = '31'
        YELLOW = '33'
        WHITE = '37'
        GREEN = '32'
        BACKGROUND_RED = '41'

        BRIGHT = '1'

        if record.levelname == 'DEBUG':
            style = ''
            background = ''
            color = GREEN

        elif record.levelname == 'INFO':
            style = ''
            background = ''
            color = WHITE

        elif record.levelname == 'WARNING':
            style = ''
            background = ''
            color = YELLOW

        elif record.levelname == 'ERROR':
            style = BRIGHT
            background = ''
            color = RED

        elif record.levelname == 'CRITICAL':
            style = BRIGHT
            background = BACKGROUND_RED
            color = WHITE

        if style:
            color = ';{}'.format(color)

        if background and color:
            background = ';{}'.format(background)

        return '\033[{}{}{}m{}\033[00m'.format(
            style,
            color,
            background,
            record_string,
        )



from copy import copy
import logging
import os

from lona.html.abstract_node import AbstractNode
from lona.types import Symbol

logger = logging.getLogger('lona.static_files')


class SORT_ORDER(Symbol):
    FRAMEWORK = Symbol('FRAMEWORK', 100)
    LIBRARY = Symbol('LIBRARY', 200)
    APPLICATION = Symbol('APPLICATION', 300)


class StaticFile:
    def __init__(self, name, path, url=None, sort_order=None, link=True,
                 enabled_by_default=True):

        self.name = name
        self.path = path
        self.url = url
        self.sort_order = sort_order or SORT_ORDER.APPLICATION
        self.link = link
        self.enabled_by_default = enabled_by_default
        self.enabled = True

        # this value is only for the object representation
        # and is meant for debugging
        self.static_url_prefix = ''

    def __repr__(self):
        return '<{}({}, sort_order={}, enabled={}, {}{} -> {})>'.format(
            self.__class__.__name__,
            self.name,
            repr(self.sort_order),
            repr(self.enabled),
            self.static_url_prefix,
            self.url,
            self.path,
        )


class StyleSheet(StaticFile):
    pass


class Script(StaticFile):
    pass


class StaticFileLoader:
    def __init__(self, server):
        self.server = server

        self.static_dirs = (self.server.settings.STATIC_DIRS)

        logger.debug('static dirs %s loaded', repr(self.static_dirs)[1:-1])

        self.discover()

    def discover_node_classes(self):
        self.node_classes = [AbstractNode]

        while True:
            count = len(self.node_classes)

            for node_class in list(self.node_classes):
                for sub_class in node_class.__subclasses__():
                    if sub_class not in self.node_classes:
                        self.node_classes.append(sub_class)

            if len(self.node_classes) == count:
                return self.node_classes

    def discover_node_static_files(self):
        self.node_stylesheets = []
        self.node_scripts = []
        self.node_static_files = []
        self.static_files = []

        # discover
        discovered_names = []

        for node_class in self.node_classes:
            if not hasattr(node_class, 'STATIC_FILES'):
                continue

            for static_file in node_class.STATIC_FILES:
                if static_file.name in discovered_names:
                    continue

                node_class_path = node_class.get_path()
                node_class_dirname = os.path.dirname(node_class_path)

                # check static file
                if not isinstance(static_file, StaticFile):
                    logger.error(
                        '%s::%s: unknown type found: %s',
                        node_class_path,
                        node_class.__name__,
                        repr(static_file),
                    )

                    continue

                # create a local copy
                static_file = copy(static_file)

                static_file.static_url_prefix = \
                    self.server.settings.STATIC_URL_PREFIX

                # patch path
                static_file_rel_path = static_file.path

                static_file_abs_path = os.path.join(
                    node_class_dirname, static_file_rel_path)

                if not os.path.exists(static_file_abs_path):
                    logger.error(
                        "%s::%s: path '%s' does not exist",
                        node_class_path,
                        node_class.__name__,
                        static_file_abs_path,
                    )

                    continue

                static_file.path = static_file_abs_path

                # patch url
                if not static_file.url:
                    url = static_file_rel_path

                else:
                    url = static_file.url

                if url.startswith('/'):
                    url = url[1:]

                static_file.url = url

                # sort static file into the right cache
                if isinstance(static_file, StyleSheet):
                    self.node_stylesheets.append(static_file)

                elif isinstance(static_file, Script):
                    self.node_scripts.append(static_file)

                else:
                    self.node_static_files.append(static_file)

                discovered_names.append(static_file.name)

        # sort
        self.node_stylesheets = sorted(
            self.node_stylesheets, key=lambda x: x.sort_order)

        self.node_scripts = sorted(
            self.node_scripts, key=lambda x: x.sort_order)

        # enable or disable static files
        self.static_files = [
            *self.node_stylesheets,
            *self.node_scripts,
            *self.node_static_files,
        ]

        for static_file in self.static_files:
            # disable static files that are not enabled by default and
            # are not configured to be enabled
            if(not static_file.enabled_by_default and
               static_file.name not in
               self.server.settings.STATIC_FILES_ENABLED):

                static_file.enabled = False

            # disable static files that are disabled explicitly
            if static_file.name in self.server.settings.STATIC_FILES_DISABLED:
                static_file.enabled = False

    def discover(self):
        logger.debug('discover node classes')

        self.discover_node_classes()
        self.discover_node_static_files()

        logger.debug('%s node classes discovered', len(self.node_classes))

        logger.debug('%s node stylesheets discovered',
                     len(self.node_stylesheets))

        logger.debug('%s node scripts discovered', len(self.node_scripts))

        # render html files
        logger.debug('rendering html files')

        # stylesheets
        self.style_tags_html = self.server.templating_engine.render_template(
            self.server.settings.STATIC_FILES_STYLE_TAGS_TEMPLATE,
            {
                'stylesheets': [i for i in self.node_stylesheets if i.enabled],
            }
        )

        # scripts
        self.script_tags_html = self.server.templating_engine.render_template(
            self.server.settings.STATIC_FILES_SCRIPT_TAGS_TEMPLATE,
            {
                'scripts': [i for i in self.node_scripts if i.enabled],
            }
        )

    def resolve_path(self, path):
        logger.debug("resolving '%s'", path)

        rel_path = path

        if rel_path.startswith('/'):
            rel_path = rel_path[1:]

        # javascript client
        client_url = self.server.settings.STATIC_FILES_CLIENT_URL

        if client_url.startswith('/'):
            client_url = client_url[1:]

        if rel_path == client_url:
            logger.debug('returning javascript client')

            return self.server.client_pre_compiler.resolve()

        # searching in static dirs
        for static_dir in self.static_dirs[::-1]:
            abs_path = os.path.join(static_dir, rel_path)

            logger.debug("trying '%s'", abs_path)

            if os.path.exists(abs_path):
                if os.path.isdir(abs_path):
                    logger.debug(
                        "'%s' is directory. resolving stopped", abs_path)

                    return

                logger.debug("returning '%s'", abs_path)

                return abs_path

        # searching in node static files
        for static_file in self.static_files:
            if not static_file.enabled:
                continue

            if static_file.url == path:
                return static_file.path

        logger.debug("'%s' was not found", path)



import os

MAX_WORKERS = 10

# routing
ROUTING_TABLE = 'lona.default_routes.routes'

# templating
CORE_TEMPLATE_DIRS = [
    os.path.join(os.path.dirname(__file__), 'templates'),
]

TEMPLATE_DIRS = []

FRONTEND_TEMPLATE = 'lona/frontend.html'
ERROR_403_TEMPLATE = 'lona/403.html'
ERROR_404_TEMPLATE = 'lona/404.html'
ERROR_500_TEMPLATE = 'lona/500.html'

TEMPLATE_EXTRA_CONTEXT = {}

# static files
STATIC_DIRS = []
STATIC_URL_PREFIX = '/static/'
STATIC_FILES_SERVE = True
STATIC_FILES_STYLE_TAGS_TEMPLATE = 'lona/style_tags.html'
STATIC_FILES_SCRIPT_TAGS_TEMPLATE = 'lona/script_tags.html'
STATIC_FILES_SYMBOLS_TEMPLATE = 'lona/lona-symbols.js'
STATIC_FILES_ENABLED = []
STATIC_FILES_DISABLED = []
STATIC_FILES_CLIENT_URL = '/lona/lona.js'

CLIENT_RECOMPILE = False

# state
SERVER_STATE_ATOMIC = True

# views
CORE_FRONTEND_VIEW = 'lona.default_views.FrontendView'
FRONTEND_VIEW = ''

# error views
CORE_ERROR_403_VIEW = 'lona.default_views.Error403View'
CORE_ERROR_404_VIEW = 'lona.default_views.Error404View'
CORE_ERROR_500_VIEW = 'lona.default_views.Error500View'

ERROR_403_VIEW = ''
ERROR_404_VIEW = ''
ERROR_500_VIEW = ''

# middlewares
CORE_MIDDLEWARES = [
    'lona.middlewares.LonaMessageMiddleware',
]

MIDDLEWARES = []

# hooks
STARTUP_HOOKS = []
SHUTDOWN_HOOKS = []

# shell
CORE_COMMANDS = [
    'lona.shell.commands.lona_server_state.LonaServerStateCommand',
    'lona.shell.commands.lona_connections.LonaConnectionsCommand',
    'lona.shell.commands.lona_settings.LonaSettingsCommand',
    'lona.shell.commands.lona_routes.LonaRoutesCommand',
    'lona.shell.commands.lona_views.LonaViewsCommand',
    'lona.shell.commands.lona_info.LonaInfoCommand',
]

COMMANDS = []



import json

from lona.types import Symbol
from lona.json import dumps

NoneType = type(None)


class PROTOCOL(Symbol):
    _INCLUDE_IN_FRONTEND_LIBRARY = True

    MESSAGE_PREFIX = Symbol('MESSAGE_PREFIX', 'lona:')


class EXIT_CODE(Symbol):
    _INCLUDE_IN_FRONTEND_LIBRARY = True

    SUCCESS = Symbol('SUCCESS', 0)
    INVALID_MESSAGE = Symbol('INVALID_MESSAGE', 1)
    INVALID_METHOD = Symbol('INVALID_METHOD', 2)


class METHOD(Symbol):
    _INCLUDE_IN_FRONTEND_LIBRARY = True

    VIEW = Symbol('VIEW', 101)
    INPUT_EVENT = Symbol('INPUT_EVENT', 102)
    REDIRECT = Symbol('REDIRECT', 201)
    HTTP_REDIRECT = Symbol('HTTP_REDIRECT', 202)
    DATA = Symbol('DATA', 203)
    VIEW_START = Symbol('VIEW_START', 204)
    VIEW_STOP = Symbol('VIEW_STOP', 205)


class INPUT_EVENT_TYPE(Symbol):
    _INCLUDE_IN_FRONTEND_LIBRARY = True

    CLICK = Symbol('CLICK', 301)
    CHANGE = Symbol('CHANGE', 302)
    SUBMIT = Symbol('SUBMIT', 303)
    CUSTOM = Symbol('CUSTOM', 304)


class NODE_TYPE(Symbol):
    _INCLUDE_IN_FRONTEND_LIBRARY = True

    NODE = Symbol('NODE', 401)
    TEXT_NODE = Symbol('TEXT_NODE', 402)
    WIDGET = Symbol('WIDGET', 403)


class DATA_TYPE(Symbol):
    _INCLUDE_IN_FRONTEND_LIBRARY = True

    HTML = Symbol('HTML', 501)
    HTML_TREE = Symbol('HTML_TREE', 502)
    HTML_UPDATE = Symbol('HTML_UPDATE', 503)


class PATCH_TYPE(Symbol):
    _INCLUDE_IN_FRONTEND_LIBRARY = True

    ID_LIST = Symbol('ID_LIST', 601)
    CLASS_LIST = Symbol('CLASS_LIST', 602)
    STYLE = Symbol('STYLE', 603)
    ATTRIBUTES = Symbol('ATTRIBUTES', 604)
    NODES = Symbol('NODES', 605)
    WIDGET_DATA = Symbol('WIDGET_DATA', 606)


class OPERATION(Symbol):
    _INCLUDE_IN_FRONTEND_LIBRARY = True

    SET = Symbol('SET', 701)
    RESET = Symbol('RESET', 702)
    ADD = Symbol('ADD', 703)
    CLEAR = Symbol('CLEAR', 704)
    INSERT = Symbol('INSERT', 705)
    REMOVE = Symbol('REMOVE', 706)


def decode_message(raw_message):
    """
    returns: (exit_code, window_id, view_runtime_id, method, payload)

    """

    def _invalid_message():
        return EXIT_CODE.INVALID_MESSAGE, None, None, None, None

    if not raw_message.startswith(PROTOCOL.MESSAGE_PREFIX.value):
        return _invalid_message()

    message = raw_message[len(PROTOCOL.MESSAGE_PREFIX.value):]

    try:
        message = json.loads(message)

    except json.JSONDecodeError:
        return _invalid_message()

    if not isinstance(message, list):
        return _invalid_message()

    if len(message) != 4:
        return _invalid_message()

    # window_id
    if not isinstance(message[0], int):
        return _invalid_message()

    # view_runtime_id
    if not isinstance(message[1], (int, NoneType)):
        return _invalid_message()

    window_id, view_runtime_id, method, payload = message

    # view
    if method == METHOD.VIEW:
        if not isinstance(payload[0], str):
            return _invalid_message()

        if not isinstance(payload[1], (dict, NoneType)):
            return _invalid_message()

        return (EXIT_CODE.SUCCESS, *message)

    # input event
    if method == METHOD.INPUT_EVENT:
        if not (isinstance(payload[0], str) or 304 > payload[0] > 300):
            return _invalid_message()

        return (EXIT_CODE.SUCCESS, *message)

    return (EXIT_CODE.INVALID_METHOD, *message)


def encode_redirect(window_id, view_runtime_id, target_url):
    return PROTOCOL.MESSAGE_PREFIX.value + dumps(
        [window_id, view_runtime_id, METHOD.REDIRECT, target_url],
    )


def encode_http_redirect(window_id, view_runtime_id, target_url):
    return PROTOCOL.MESSAGE_PREFIX.value + dumps(
        [window_id, view_runtime_id, METHOD.HTTP_REDIRECT, target_url],
    )


def encode_data(window_id, view_runtime_id, title, data):
    return PROTOCOL.MESSAGE_PREFIX.value + dumps(
        [window_id, view_runtime_id, METHOD.DATA, [title, data]],
    )


def encode_view_start(window_id, view_runtime_id):
    return PROTOCOL.MESSAGE_PREFIX.value + dumps(
        [window_id, view_runtime_id, METHOD.VIEW_START, None],
    )


def encode_view_stop(window_id, view_runtime_id):
    return PROTOCOL.MESSAGE_PREFIX.value + dumps(
        [window_id, view_runtime_id, METHOD.VIEW_STOP, view_runtime_id, None],
    )



from threading import RLock
import logging

from lona.protocol import DATA_TYPE
from lona.imports import acquire

AbstractNode = None
Widget = None
Node = None

_node_classes_setup = False

logger = logging.getLogger('lona.html.document')


def _setup_node_classes():
    # this is necessary to prevent import loops between the Document
    # class and the AbstractNode class and its subclasses

    global AbstractNode, TextNode, Widget, Node, _node_classes_setup

    if _node_classes_setup:
        return

    AbstractNode = acquire('lona.html.abstract_node.AbstractNode')
    Widget = acquire('lona.html.widget.Widget')
    Node = acquire('lona.html.node.Node')

    _node_classes_setup = True


class Document:
    def __init__(self, default_document=False):
        self.is_default_document = default_document

        self._lock = RLock()
        self.html = None

        if not self.is_default_document:
            _setup_node_classes()

    @property
    def lock(self):
        return self._lock

    def __repr__(self):
        if self.is_default_document:
            return '<DefaultDocument>'

        return '<Document>'

    # html ####################################################################
    def get_node(self, node_id, widget_id=None):
        if isinstance(node_id, Node):
            node_id = node_id.id

        value = [None, [], ]
        widget = [None]
        widget_path = []

        def iter_nodes(node):
            if isinstance(node, Widget):
                widget_path.append(node)

                if widget_id and node._id == widget_id:
                    widget[0] = node

            if isinstance(node, (Node, Widget)):
                if node._id == node_id:
                    value[0] = node
                    value[1].extend(widget_path)

                    return

                if node.nodes:
                    for i in node.nodes:
                        iter_nodes(i)

            if isinstance(node, Widget):
                widget_path.pop()

        with self.lock:
            iter_nodes(self.html)

        if widget[0]:
            value[1].append(widget[0])

        return tuple(value)

    def _has_changes(self):
        def has_changes(node):
            if node._has_changes():
                return True

            if hasattr(node, 'nodes'):
                for sub_node in node.nodes:
                    if has_changes(sub_node):
                        return True

            return False

        return has_changes(self.html)

    def _collect_changes(self):
        changes = []
        changed_widgets = []
        widget_path = []

        # find changes and changed widgets
        def add_changes(node):
            node_is_widget = isinstance(node, Widget)

            if node_is_widget:
                widget_path.append(node._id)

            # changed widgets
            if node_is_widget:
                if node.nodes._has_changes():
                    changed_widgets.extend(widget_path)

            elif node._has_changes():
                changed_widgets.extend(widget_path)

            # changes
            if node._has_changes():
                changes.extend(node._get_changes())

            if hasattr(node, 'nodes'):
                for sub_node in node.nodes:
                    add_changes(sub_node)

            if node_is_widget:
                widget_path.remove(node._id)

        add_changes(self.html)

        # sort changes by timestamp
        changes = sorted(changes, key=lambda x: x[0])

        # remove timestamps
        cleaned_changes = []

        for change in changes:
            cleaned_changes.append(change[1:])

        # clean list of changed widgets
        cleaned_changed_widgets = []

        for widget_id in changed_widgets[::-1]:
            if widget_id not in cleaned_changed_widgets:
                cleaned_changed_widgets.append(widget_id)

        return [cleaned_changes, cleaned_changed_widgets]

    def serialize(self):
        if not self.html:
            return self.apply('')

        return DATA_TYPE.HTML_TREE, self.html._serialize()

    def apply(self, html):
        if isinstance(html, str) and html is self.html:
            return

        # HTML update
        elif html is self.html:
            if not self._has_changes():
                return

            changes = self._collect_changes()
            self.html._clear_changes()

            return DATA_TYPE.HTML_UPDATE, changes

        # HTML
        else:
            if hasattr(self.html, 'document'):
                self.html.document = None

            # node tree
            if isinstance(html, AbstractNode):
                self.html = html

                self.html.document = self
                self.html._clear_changes()

                return self.serialize()

            # HTML string
            self.html = str(html)

            return DATA_TYPE.HTML, html



import shutil
import os

from aiohttp.web import Application

from lona.server import LonaServer


def collect_static(args):
    def _print(*print_args, **print_kwargs):
        if args.silent:
            return

        print(*print_args, **print_kwargs)

    def _rm(path):
        _print('rm -rf {}'.format(path))

        if args.dry_run:
            return

        if os.path.isdir(path):
            shutil.rmtree(path)

        else:
            os.unlink(path)

    def _mkdir(path):
        _print('mkdir -p {}'.format(path))

        if args.dry_run:
            return

        try:
            os.makedirs(path)

        except FileExistsError:
            pass

    def _cp(source, destination):
        _print('cp {} {}'.format(source, destination))

        if args.dry_run:
            return

        shutil.copy(source, destination)

    # setup server
    server = LonaServer(
        app=Application(),
        project_root=args.project_root,
        settings_paths=args.settings,
        settings_pre_overrides=args.settings_pre_overrides,
        settings_post_overrides=args.settings_post_overrides,
    )

    # check destination
    if not os.path.exists(args.destination):
        exit("'{}' does not exist".format(args.destination))

    if not os.path.isdir(args.destination):
        exit("'{}' is no directory".format(args.destination))

    # clean
    if args.clean:
        for name in os.listdir(args.destination):
            path = os.path.join(args.destination, name)
            _rm(path)

    # copy node static files
    for static_file in server.static_file_loader.static_files[::-1]:
        if not static_file.enabled:
            continue

        destination = os.path.join(
            args.destination,
            static_file.url,
        )

        _cp(static_file.path, destination)

    # copy static files from static directories
    for static_dir in server.static_file_loader.static_dirs[::-1]:
        for name in os.listdir(static_dir):
            source = os.path.join(static_dir, name)
            destination = os.path.join(args.destination, name)

            _cp(source, destination)

    # copy client
    client_url = server.settings.STATIC_FILES_CLIENT_URL

    if client_url.startswith('/'):
        client_url = client_url[1:]

    destination = os.path.join(
        args.destination,
        client_url,
    )

    _mkdir(os.path.dirname(destination))
    _cp(server.client_pre_compiler.resolve(), destination)



import inspect
import time

from lona.html.document import Document
from lona.html.selector import Selector

_default_document = Document(default_document=True)


class AbstractNode:
    @classmethod
    def get_path(cls):
        if cls.__module__.endswith('.py'):
            return cls.__module__

        return inspect.getfile(cls)

    @classmethod
    def gen_id(cls):
        return time.monotonic_ns()

    @property
    def parent(self):
        if not hasattr(self, '_parent'):
            self._parent = None

        return self._parent

    @parent.setter
    def parent(self, value):
        self._parent = value

    @property
    def root(self):
        node = self

        while True:
            if node.parent is None:
                break

            node = node.parent

        return node

    @property
    def document(self):
        return getattr(
            self.root, '_document', _default_document) or _default_document

    @document.setter
    def document(self, value):
        self._document = value

    @property
    def lock(self):
        return self.document.lock

    def __copy__(self, *args, **kwargs):
        raise RuntimeError('copy is not supported')

    def __deepcopy__(self, memo):
        raise RuntimeError('deepcopy is not supported')

    # queries #################################################################
    def iter_nodes(self, node=None):
        node = node or self

        if hasattr(node, 'nodes'):
            for child in node.nodes:
                yield child

                for sub_child in self.iter_nodes(child):
                    yield sub_child

    def query_selector(self, raw_selector_string):
        selector = Selector(raw_selector_string)

        with self.lock:
            for node in self.iter_nodes():
                if selector.match(node):
                    return node

    def query_selector_all(self, raw_selector_string):
        selector = Selector(raw_selector_string)
        nodes = []

        with self.lock:
            for node in self.iter_nodes():
                if selector.match(node):
                    nodes.append(node)

            return nodes



from time import monotonic_ns

from lona.protocol import OPERATION, PATCH_TYPE


class AttributeDict:
    PATCH_TYPE = PATCH_TYPE.ATTRIBUTES

    def __init__(self, node, *args, **kwargs):
        self._node = node
        self._attributes = dict(*args, **kwargs)
        self._changes = []

    # dict helper #############################################################
    def keys(self):
        with self._node.document.lock:
            return self._attributes.keys()

    def items(self):
        with self._node.document.lock:
            return self._attributes.items()

    def pop(self, name):
        with self._node.document.lock:
            attribute = self._attributes.pop(name)

            self._changes.append([
                monotonic_ns(),
                self._node._id,
                self.PATCH_TYPE,
                OPERATION.REMOVE,
                name,
            ])

            return attribute

    def clear(self):
        with self._node.document.lock:
            if not self._attributes:
                return

            self._attributes.clear()
            self._changes.append([
                monotonic_ns(),
                self._node._id,
                self.PATCH_TYPE,
                OPERATION.CLEAR,
            ])

    def get(self, *args, **kwargs):
        with self._node.document.lock:
            return self._attributes.get(*args, **kwargs)

    def update(self, value):
        if not isinstance(value, dict):
            raise ValueError('dict required')

        with self._node.document.lock:
            for key, value in value.items():
                self[key] = value

    def __getitem__(self, name):
        with self._node.document.lock:
            return self._attributes[name]

    def __setitem__(self, name, value):
        if not isinstance(value, (int, bool, float, str)):
            raise ValueError('unsupported type: {}'.format(type(value)))

        if name in ('id', 'class', 'style'):
            raise RuntimeError(
                "Node.attributes['{}'] is not supported. "
                'Use Node.{}{} instead.'.format(
                    name, name, '_list' if name != 'style' else '')
            )

        with self._node.document.lock:
            if name in self._attributes and self._attributes[name] == value:
                return

            self._attributes[name] = value

            self._changes.append([
                monotonic_ns(),
                self._node._id,
                self.PATCH_TYPE,
                OPERATION.SET,
                name,
                value,
            ])

    def __delitem__(self, name):
        with self._node.document.lock:
            del self._attributes[name]

            self._changes.append([
                monotonic_ns(),
                self._node._id,
                self.PATCH_TYPE,
                OPERATION.REMOVE,
                name,
            ])

    def __iter__(self):
        with self._node.document.lock:
            return self._attributes.__iter__()

    def __bool__(self):
        with self._node.document.lock:
            return bool(self._attributes)

    # serialisation ###########################################################
    def _reset(self, value):
        if not isinstance(value, dict):
            raise ValueError('unsupported type')

        for k, v in value.items():
            if not isinstance(v, (int, bool, float, str)):
                raise ValueError('unsupported type')

            if k in ('id', 'class', 'style'):
                raise RuntimeError(
                    "Node.attributes['{}'] is not supported. "
                    'Use Node.{}{} instead.'.format(
                        k, k, 'list' if k != 'style' else '')
                )

        with self._node.document.lock:
            self._attributes = value

            self._changes.append([
                monotonic_ns(),
                self._node._id,
                self.PATCH_TYPE,
                OPERATION.RESET,
                dict(value),
            ])

    def _has_changes(self):
        return bool(self._changes)

    def _get_changes(self):
        return list(self._changes)

    def _clear_changes(self):
        return self._changes.clear()

    def _serialize(self):
        return dict(self._attributes)

    # string representation ###################################################
    def to_attribute_string(self, skip_value=False):
        with self._node.document.lock:
            string = []

            for key, value in self._attributes.items():
                if skip_value and key == 'value':
                    continue

                string.append('{}="{}"'.format(key, value))

            return ' '.join(string)

    def to_sub_attribute_string(self):
        with self._node.document.lock:
            string = []

            for key, value in self._attributes.items():
                string.append('{}: {}'.format(key, value))

            return '; '.join(string)

    def __repr__(self):
        return '<AttributeDict({})>'.format(repr(self._attributes))


class StyleDict(AttributeDict):
    PATCH_TYPE = PATCH_TYPE.STYLE

    def __repr__(self):
        return '<StyleDict({})>'.format(repr(self._attributes))



from tempfile import TemporaryDirectory
import logging
import os

from jinja2 import Environment, FileSystemLoader

from lona.types import Symbol
from lona.json import dumps

SOURCE_ROOT = os.path.join(os.path.dirname(__file__), 'client')

logger = logging.getLogger('lona.client_pre_compiler')


class ClientPreCompiler:
    def __init__(self, server):
        self.server = server

        self.tmp_dir = TemporaryDirectory()

        self.jinja2_env = Environment(
            loader=FileSystemLoader(SOURCE_ROOT),
        )

        self.compile()

    def _get_path(self):
        return os.path.join(self.tmp_dir.name, 'lona.js')

    def compile(self):
        logger.debug('pre compiling client')

        try:
            path = self._get_path()
            template = self.jinja2_env.get_template('lona.js')

            template_context = {
                'symbols': dumps(Symbol.dump_symbol_classes()),
            }

            file_content = template.render(
                **template_context,
            )

            with open(path, 'w+') as f:
                f.write(file_content)
                f.close()

        except Exception:
            logger.error(
                'exception raised while pre compiling js client',
                exc_info=True,
            )

    def resolve(self):
        if self.server.settings.CLIENT_RECOMPILE:
            self.compile()

        return self._get_path()

    def __repr__(self):
        return '<ClientPreCompiler({})>'.format(self._get_path())
